"""正規化キーによる財務数値の抽出ユーティリティ。

会計基準（J-GAAP / IFRS / US-GAAP / JMIS）を意識せず、canonical key で
SummaryOfBusinessResults および PL/BS/CF 本体から値を取り出す。

v0.3.0 設計:
    ``Statements`` を渡すと ``_items`` 全体をパイプラインマッパーで 1 パス走査する。
    デフォルトパイプライン
    ``[summary_mapper, statement_mapper, definition_mapper(), calc_mapper()]``
    で名寄せを行う。

使用例::

    from edinet.financial.extract import extract_values, extracted_to_dict
    from edinet.financial.standards.canonical_keys import CK

    # Statements → Summary + PL/BS/CF から当期・連結の値を抽出
    result = extract_values(stmts, [CK.REVENUE, CK.OPERATING_INCOME],
                            period="current", consolidated=True)

    # pandas 連携
    row = extracted_to_dict(result)
"""

from __future__ import annotations

import warnings
from collections.abc import Sequence
from dataclasses import dataclass
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from edinet.financial.statements import Statements

from edinet.exceptions import EdinetWarning
from edinet.financial.mapper import (
    ConceptMapper,
    MapperContext,
    calc_mapper,
    definition_mapper,
    statement_mapper,
    summary_mapper,
)
from edinet.financial.standards.canonical_keys import CK
from edinet.models.financial import LineItem
from edinet.financial.statements import (
    _is_consolidated as _check_consolidated,
    _is_non_consolidated as _check_non_consolidated,
)

__all__ = ["ExtractedValue", "extract_values", "extracted_to_dict"]

_CK_MEMBERS: frozenset[str] = frozenset(CK)
"""CK enum メンバー集合。typo 警告の O(1) ルックアップ用。"""


@dataclass(frozen=True, slots=True)
class ExtractedValue:
    """正規化キーで抽出された財務数値。

    Attributes:
        canonical_key: 正規化キー（例: ``"revenue"``）。
        value: 抽出された値。数値の場合は ``Decimal``、テキストの場合は
            ``str``、nil または欠損の場合は ``None``。
        item: 元の ``LineItem``。``source_line``、``label_ja``、
            ``namespace_uri`` 等のトレーサビリティ情報を含む。
        mapper_name: 値を採用したマッパー関数の名前。
            ``getattr(mapper_fn, "__name__", None)`` で自動取得される。
            例: ``"summary_mapper"``、``"statement_mapper"``、
            ``"dict_mapper(3 entries)"``。
    """

    canonical_key: str
    value: Decimal | str | None
    item: LineItem
    mapper_name: str | None


# ---------------------------------------------------------------------------
# 内部ヘルパー
# ---------------------------------------------------------------------------


def _resolve_periods(
    period: Literal["current", "prior"] | None,
    dei: Any | None,
) -> tuple[Any | None, Any | None]:
    """period 文字列を DurationPeriod/InstantPeriod に解決する。

    Args:
        period: 期間フィルタ。
        dei: DEI 情報。

    Returns:
        ``(target_duration, target_instant)`` のタプル。
    """
    if period is None:
        return None, None

    if dei is None:
        warnings.warn(
            f"period={period!r} が指定されましたが DEI 情報がないため"
            f"期間フィルタをスキップします",
            EdinetWarning,
            stacklevel=4,
        )
        return None, None

    from edinet.financial.dimensions.period_variants import classify_periods

    pc = classify_periods(dei)
    if period == "current":
        return pc.current_duration, pc.current_instant
    # period == "prior"
    return pc.prior_duration, pc.prior_instant


def _filter_item(
    item: LineItem,
    *,
    period: Literal["current", "prior"] | None,
    target_duration: Any | None,
    target_instant: Any | None,
    consolidated: bool | None,
) -> bool:
    """期間・連結フィルタを適用し、通過すれば True を返す。"""
    from edinet.xbrl.contexts import DurationPeriod, InstantPeriod

    # 期間フィルタ
    if period is not None and (target_duration is not None or target_instant is not None):
        if isinstance(item.period, DurationPeriod):
            if target_duration is not None and item.period != target_duration:
                return False
        elif isinstance(item.period, InstantPeriod):
            if target_instant is not None and item.period != target_instant:
                return False

    # 連結フィルタ
    if consolidated is not None:
        if consolidated and not _check_consolidated(item):
            return False
        if not consolidated and not _check_non_consolidated(item):
            return False

    return True


# ---------------------------------------------------------------------------
# 公開 API
# ---------------------------------------------------------------------------


def extract_values(
    source: Statements,
    keys: Sequence[str] | None = None,
    *,
    period: Literal["current", "prior"] | None = None,
    consolidated: bool | None = None,
    mapper: ConceptMapper | Sequence[ConceptMapper] | None = None,
) -> dict[str, ExtractedValue | None]:
    """正規化キーで財務データから値を抽出する。

    ``Statements`` を渡すと ``_items`` 全体をパイプラインマッパーで
    1 パス走査する。

    Args:
        source: 抽出対象の ``Statements``。
        keys: 抽出する正規化キーのシーケンス。
            ``CK`` enum または文字列で指定可能。
            ``None`` の場合は全マッピング可能科目を抽出する。
        period: 期間フィルタ。
            ``"current"`` で当期、``"prior"`` で前期。
            ``None`` の場合は全期間から先頭マッチ。
        consolidated: 連結フィルタ。
            ``True`` で連結、``False`` で個別。
            ``None`` の場合は全て。
        mapper: マッパーまたはマッパーのシーケンス。
            ``None``（デフォルト）の場合は
            ``[summary_mapper, statement_mapper, definition_mapper(), calc_mapper()]``。
            単一 callable の場合は ``[callable]`` と同義。

    Returns:
        ``{canonical_key: ExtractedValue | None}`` の辞書。
        ``keys`` で指定されたキーが見つからない場合は ``None``。
        ``keys=None`` の場合は見つかった科目のみを含む。

    Raises:
        TypeError: ``Statements`` 以外を渡した場合。

    Example:
        >>> from edinet.financial.standards.canonical_keys import CK
        >>> result = extract_values(stmts, [CK.REVENUE, CK.OPERATING_INCOME],
        ...                         period="current", consolidated=True)
        >>> result["operating_income"].value
        Decimal('500000000')
    """
    from edinet.financial.statements import Statements as _Statements

    if not isinstance(source, _Statements):
        raise TypeError(
            f"Statements を渡してください（got {type(source).__name__}）"
        )

    # マッパーパイプライン解決
    if mapper is None:
        pipeline: list[ConceptMapper] = [
            summary_mapper, statement_mapper,
            definition_mapper(), calc_mapper(),
        ]
    elif callable(mapper):
        pipeline = [mapper]
    else:
        pipeline = list(mapper)

    # MapperContext 構築
    from edinet.xbrl.taxonomy.custom import _build_parent_index

    ctx = MapperContext(
        dei=source.dei,
        detected_standard=source.detected_standard,
        industry_code=source.industry_code,
        definition_parent_index=_build_parent_index(
            source.definition_linkbase,
        ),
        calculation_linkbase=source.calculation_linkbase,
    )

    # 期間解決（1回のみ）
    target_duration, target_instant = _resolve_periods(
        period, source.dei,
    )

    # キーフィルタ
    target_keys: set[str] | None = (
        {str(k) for k in keys} if keys is not None else None
    )

    # 1パスマッパーループ
    # ck_to_item: {canonical_key: (LineItem, pipeline_position, mapper_name)}
    ck_to_item: dict[str, tuple[LineItem, int, str | None]] = {}

    for item in source:
        if not _filter_item(
            item,
            period=period,
            target_duration=target_duration,
            target_instant=target_instant,
            consolidated=consolidated,
        ):
            continue

        for idx, mapper_fn in enumerate(pipeline):
            ck = mapper_fn(item, ctx)
            if ck is None:
                continue
            if target_keys is not None and ck not in target_keys:
                continue
            # CK typo 警告（keys=None 時のみ）
            if target_keys is None and ck not in _CK_MEMBERS:
                warnings.warn(
                    f"マッパーが未知の CK '{ck}' を返しました。"
                    f"タイポの可能性があります。",
                    UserWarning,
                    stacklevel=2,
                )
            name = getattr(mapper_fn, "__name__", None)
            existing = ck_to_item.get(ck)
            if existing is None or idx < existing[1]:
                ck_to_item[ck] = (item, idx, name)
            break  # 1 item に対して最初にマッチしたマッパーを採用

    # 結果構築
    if keys is None:
        return {
            ck: ExtractedValue(
                canonical_key=ck, value=item.value, item=item,
                mapper_name=name,
            )
            for ck, (item, _idx, name) in ck_to_item.items()
        }

    result: dict[str, ExtractedValue | None] = {}
    for key in keys:
        key_str = str(key)
        entry = ck_to_item.get(key_str)
        if entry is not None:
            item, _idx, name = entry
            result[key_str] = ExtractedValue(
                canonical_key=key_str, value=item.value, item=item,
                mapper_name=name,
            )
        else:
            result[key_str] = None
    return result


def extracted_to_dict(
    *extracted_dicts: dict[str, ExtractedValue | None],
) -> dict[str, Decimal | str | None]:
    """``extract_values()`` の結果を ``{key: value}`` 辞書に変換する。

    複数の辞書を渡すとマージされる。値が見つかったキーが優先され、
    ``None`` で上書きされることはない。

    Args:
        *extracted_dicts: ``extract_values()`` の戻り値（1つ以上）。

    Returns:
        ``{canonical_key: value}`` の辞書。

    Example:
        >>> row = extracted_to_dict(
        ...     extract_values(stmts, [CK.REVENUE]),
        ...     extract_values(stmts, [CK.TOTAL_ASSETS]),
        ... )
    """
    result: dict[str, Decimal | str | None] = {}
    for extracted in extracted_dicts:
        for k, ev in extracted.items():
            if ev is not None:
                result[k] = ev.value
            elif k not in result:
                result[k] = None
    return result



"""パイプラインマッパー — concept → canonical key の名寄せロジック。

``extract_values()`` が使用するマッパーの型定義・デフォルト実装・ヘルパーを提供する。

設計原則:
    - マッパーは ``Callable[[LineItem, MapperContext], str | None]`` シグネチャ
    - ``str`` を返すとマッチ（canonical key）、``None`` で次のマッパーへ
    - パイプライン（リスト）の先頭マッパーほど高優先

デフォルトパイプライン::

    [summary_mapper, statement_mapper, definition_mapper(), calc_mapper()]

    1. summary_mapper: SummaryOfBusinessResults（経営指標）
    2. statement_mapper: PL/BS/CF の辞書引き
    3. definition_mapper(): Definition Linkbase の general-special で
       独自科目 → 標準科目に遡上
    4. calc_mapper(): Calculation Linkbase の summation-item で
       独自科目 → 親標準科目に遡上

典型的な使用例::

    from edinet.financial.mapper import summary_mapper, statement_mapper, dict_mapper

    # デフォルト: [summary_mapper, statement_mapper, definition_mapper(), calc_mapper()]
    # カスタム辞書を最優先に追加
    my_mapper = dict_mapper({"MyRevenue": "revenue"}, name="my_rules")
    pipeline = [my_mapper, summary_mapper, statement_mapper]

    # カスタム lookup を使う例（独自辞書でリンクベース解決）
    my_map = {"NetSales": "売上高", "OperatingIncome": "営業利益"}
    pipeline = [
        dict_mapper(my_map),
        definition_mapper(lookup=my_map.get),
        calc_mapper(lookup=my_map.get),
    ]
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from edinet.financial.standards.statement_mappings import (
    lookup_statement_exact,
    lookup_statement_normalized,
)
from edinet.financial.standards.summary_mappings import lookup_summary

if TYPE_CHECKING:
    from edinet.financial.standards.detect import DetectedStandard
    from edinet.models.financial import LineItem
    from edinet.xbrl.dei import DEI
    from edinet.xbrl.linkbase.calculation import CalculationLinkbase

__all__ = [
    "ConceptMapper",
    "MapperContext",
    "calc_mapper",
    "definition_mapper",
    "dict_mapper",
    "statement_mapper",
    "summary_mapper",
]

ConceptMapper = Callable[["LineItem", "MapperContext"], str | None]
"""concept → canonical key マッパーの型。

Args:
    item: 走査中の LineItem。
    ctx: マッパーコンテキスト（DEI・会計基準・業種コード・リンクベース）。

Returns:
    マッチした canonical key（``str``）。マッチしない場合は ``None``。
"""


@dataclass(frozen=True, slots=True)
class MapperContext:
    """マッパーに渡されるコンテキスト情報。

    ``extract_values()`` が ``Statements`` から自動構築する。
    ユニットテスト時は直接構築可能。

    Attributes:
        dei: DEI（Document and Entity Information）。
            会計基準、企業名、EDINET コード、決算期、報告期間種別等を含む。
            DEI が取得できなかった場合は ``None``。
        detected_standard: 判別された会計基準（J-GAAP / IFRS / US-GAAP / JMIS）。
            判別方法（DEI / namespace）と詳細度（DETAILED / BLOCK_ONLY）を含む。
        industry_code: 業種コード（例: ``"bk1"``）。
            ``None`` は一般事業会社。
        definition_parent_index: Definition Linkbase の general-special arcrole
            による逆引きインデックス。``{独自科目local_name: 標準祖先local_name}``。
            ``definition_mapper`` が使用する。``None`` の場合は
            ``definition_mapper`` は常に ``None`` を返す。
        calculation_linkbase: 提出者の Calculation Linkbase。
            ``calc_mapper`` が ``ancestors_of()`` で祖先を辿る。
            ``None`` の場合は ``calc_mapper`` は常に ``None`` を返す。
    """

    dei: DEI | None
    detected_standard: DetectedStandard | None
    industry_code: str | None
    definition_parent_index: dict[str, str] = field(default_factory=dict)
    calculation_linkbase: CalculationLinkbase | None = None


def summary_mapper(item: LineItem, ctx: MapperContext) -> str | None:
    """SummaryOfBusinessResults（経営成績の概況）から名寄せするマッパー。

    開示府令で記載が義務付けられた標準コンセプトを対象とする。
    全社・全会計基準で安定して取得できる。

    Args:
        item: 走査中の LineItem。
        ctx: マッパーコンテキスト（未使用だがシグネチャ統一のため受け取る）。

    Returns:
        マッチした canonical key。マッチしない場合は ``None``。
    """
    return lookup_summary(item.local_name)


def statement_mapper(item: LineItem, ctx: MapperContext) -> str | None:
    """財務諸表本体（PL/BS/CF）から名寄せするマッパー。

    summary で取得できなかった項目を補完する。
    完全一致 → サフィックス除去後一致の順で試行する。

    Args:
        item: 走査中の LineItem。
        ctx: マッパーコンテキスト（未使用だがシグネチャ統一のため受け取る）。

    Returns:
        マッチした canonical key。マッチしない場合は ``None``。
    """
    ck = lookup_statement_exact(item.local_name)
    if ck is not None:
        return ck
    return lookup_statement_normalized(item.local_name)


def _default_statement_lookup(concept_name: str) -> str | None:
    """組み込み statement_mappings での2段 lookup。"""
    ck = lookup_statement_exact(concept_name)
    if ck is not None:
        return ck
    return lookup_statement_normalized(concept_name)


def definition_mapper(
    lookup: Callable[[str], str | None] | None = None,
) -> ConceptMapper:
    """Definition Linkbase の general-special で標準概念に遡上し CK を返すマッパーを生成する。

    提出者独自の科目名を Definition Linkbase の general-special arcrole を
    辿り、最も近い標準タクソノミの祖先概念を見つけて CK に変換する。

    事前に ``_build_parent_index()`` で構築した逆引きインデックスを使用するため、
    マッパー呼び出し時のコストは O(1) の辞書参照 + lookup のみ。

    Args:
        lookup: 祖先 concept 名 → canonical key の名寄せ関数。
            ``None`` の場合は組み込み statement_mappings を使用する。

    Returns:
        ``ConceptMapper`` として使える callable。
    """
    _lookup = lookup or _default_statement_lookup

    def _mapper(item: LineItem, ctx: MapperContext) -> str | None:
        if not ctx.definition_parent_index:
            return None
        ancestor = ctx.definition_parent_index.get(item.local_name)
        if ancestor is None:
            return None
        return _lookup(ancestor)

    _mapper.__name__ = "definition_mapper"
    _mapper.__qualname__ = "definition_mapper"
    return _mapper


def calc_mapper(
    lookup: Callable[[str], str | None] | None = None,
) -> ConceptMapper:
    """Calculation Linkbase の summation-item で親標準概念に遡上し CK を返すマッパーを生成する。

    提出者独自の科目名を Calculation Linkbase の親子関係（summation-item
    arcrole）を辿り、祖先に標準概念が見つかれば CK に変換する。

    全 role_uri を走査し、最初に CK が見つかった時点で返す。

    Args:
        lookup: 祖先 concept 名 → canonical key の名寄せ関数。
            ``None`` の場合は組み込み statement_mappings を使用する。

    Returns:
        ``ConceptMapper`` として使える callable。
    """
    _lookup = lookup or _default_statement_lookup

    def _mapper(item: LineItem, ctx: MapperContext) -> str | None:
        if ctx.calculation_linkbase is None:
            return None
        for role_uri in ctx.calculation_linkbase.role_uris:
            ancestors = ctx.calculation_linkbase.ancestors_of(
                item.local_name, role_uri=role_uri,
            )
            for ancestor in ancestors:
                ck = _lookup(ancestor)
                if ck is not None:
                    return ck
        return None

    _mapper.__name__ = "calc_mapper"
    _mapper.__qualname__ = "calc_mapper"
    return _mapper


def dict_mapper(
    mapping: dict[str, str],
    *,
    name: str | None = None,
) -> ConceptMapper:
    """辞書ベースのマッパーを生成する。

    Excel/CSV で管理しているマッピングテーブルをそのまま利用可能。

    Args:
        mapping: ``{concept_local_name: canonical_key}`` の辞書。
        name: マッパー名（``mapper_name`` に反映される）。
            未指定の場合は ``"dict_mapper(N entries)"``。

    Returns:
        ``ConceptMapper`` として使える callable。

    Example:
        >>> m = dict_mapper({"MyRevenue": "revenue"}, name="my_rules")
        >>> m.__name__
        'my_rules'
    """

    def _mapper(item: LineItem, ctx: MapperContext) -> str | None:
        return mapping.get(item.local_name)

    _mapper.__name__ = name or f"dict_mapper({len(mapping)} entries)"
    _mapper.__qualname__ = _mapper.__name__
    return _mapper



from edinet.api._errors import raise_for_api_error_response as raise_for_api_error_response
from typing import Any

def get_documents(date: str, *, include_details: bool = True) -> dict[str, Any]:
    '''指定日の提出書類一覧を取得する。

    Args:
        date:
            ファイル日付（YYYY-MM-DD 形式）。
            当日以前で、10年を経過していない日付。
            土日祝日も指定可能。
        include_details:
            True  → type=2「提出書類一覧及びメタデータ」を取得（デフォルト）
            False → type=1「メタデータのみ」を取得

    Returns:
        EDINET API のレスポンス JSON（dict）。

    Raises:
        EdinetConfigError: API キーが未設定。
        EdinetAPIError: EDINET API がエラーを返した（HTTP エラーまたは
                        レスポンス JSON の metadata.status が "200" 以外）。

    Examples:
        >>> result = get_documents("2026-02-07")
        >>> result["metadata"]["resultset"]["count"]
        42
        >>> result["results"][0]["docID"]
        \'S100XXXX\'
    '''
async def aget_documents(date: str, *, include_details: bool = True) -> dict[str, Any]:
    '''指定日の提出書類一覧を非同期で取得する。

    Args:
        date:
            ファイル日付（YYYY-MM-DD 形式）。
            当日以前で、10年を経過していない日付。
            土日祝日も指定可能。
        include_details:
            True  → type=2「提出書類一覧及びメタデータ」を取得（デフォルト）
            False → type=1「メタデータのみ」を取得

    Returns:
        EDINET API のレスポンス JSON（dict）。

    Raises:
        EdinetConfigError: API キーが未設定。
        EdinetAPIError: EDINET API がエラーを返した（HTTP エラーまたは
                        レスポンス JSON の metadata.status が "200" 以外）。
    '''



from _typeshed import Incomplete
from dataclasses import dataclass
from edinet.exceptions import EdinetParseError as EdinetParseError, EdinetWarning as EdinetWarning
from edinet.xbrl._namespaces import NS_LINK as NS_LINK, NS_XLINK as NS_XLINK
from typing import Literal

logger: Incomplete
ARCROLE_SUMMATION: str

@dataclass(frozen=True, slots=True)
class CalculationArc:
    '''計算リンクベースの 1 本のアーク。

    親 concept から子 concept への加減算関係を表す。

    Attributes:
        parent: 親 concept のローカル名（例: ``"GrossProfit"``）。
        child: 子 concept のローカル名（例: ``"NetSales"``）。
        parent_href: 親 concept の xlink:href（元の値）。
        child_href: 子 concept の xlink:href（元の値）。
        weight: 加減算方向。``1`` が加算、``-1`` が減算。
        order: 同一親の下での表示順序。
        role_uri: 所属ロール URI。
    '''
    parent: str
    child: str
    parent_href: str
    child_href: str
    weight: Literal[1, -1]
    order: float
    role_uri: str

@dataclass(frozen=True, slots=True)
class CalculationTree:
    """1 つの role URI 内の計算木。

    Attributes:
        role_uri: ロール URI。
        arcs: 全アークのタプル（パース順保持）。
        roots: 親のみで子でない concept のタプル（XML 初出順）。
    """
    role_uri: str
    arcs: tuple[CalculationArc, ...]
    roots: tuple[str, ...]

@dataclass(frozen=True, slots=True)
class CalculationLinkbase:
    """計算リンクベースのパース結果全体。

    Attributes:
        source_path: ソースファイルパス（エラーメッセージ用）。
        trees: role_uri → CalculationTree の辞書。
    """
    source_path: str | None
    trees: dict[str, CalculationTree]
    def __post_init__(self) -> None:
        """内部インデックスを構築する。"""
    @property
    def role_uris(self) -> tuple[str, ...]:
        """全ロール URI のタプルを返す。

        Returns:
            ロール URI のタプル。
        """
    def get_tree(self, role_uri: str) -> CalculationTree | None:
        """指定ロールの計算木を返す。

        Args:
            role_uri: ロール URI。

        Returns:
            CalculationTree。存在しない場合は ``None``。
        """
    def children_of(self, parent: str, *, role_uri: str | None = None) -> tuple[CalculationArc, ...]:
        """指定 concept の子アークを返す。

        Args:
            parent: 親 concept のローカル名。
            role_uri: ロール URI。``None`` の場合は全ロールを横断する。

        Returns:
            子アークのタプル。role_uri 指定時は order 昇順、
            ``None`` 時は ``(role_uri, order)`` 昇順。
        """
    def parent_of(self, child: str, *, role_uri: str | None = None) -> tuple[CalculationArc, ...]:
        """指定 concept の親アークを返す。

        Args:
            child: 子 concept のローカル名。
            role_uri: ロール URI。``None`` の場合は全ロールを横断する。

        Returns:
            親アークのタプル。
        """
    def ancestors_of(self, concept: str, *, role_uri: str) -> tuple[str, ...]:
        """指定 concept の祖先を根まで辿って返す。

        ``_parent_index`` を辿り、複数親がある場合は先頭を辿る。
        循環は ``visited`` で防御する。

        Args:
            concept: 起点 concept のローカル名。
            role_uri: ロール URI。

        Returns:
            祖先 concept のタプル（近い順）。ルート concept が末尾。
        """

def parse_calculation_linkbase(xml_bytes: bytes, *, source_path: str | None = None) -> CalculationLinkbase:
    """Calculation Linkbase の XML bytes をパースする。

    Args:
        xml_bytes: ``_cal.xml`` の bytes。
        source_path: エラーメッセージに含めるソースパス（任意）。

    Returns:
        パース結果の CalculationLinkbase。

    Raises:
        EdinetParseError: XML パースに失敗した場合。
    """



from _typeshed import Incomplete
from datetime import date as DateType
from edinet._validators import normalize_edinet_code as normalize_edinet_code
from edinet.models.doc_types import DocType as DocType
from edinet.models.edinet_code import EdinetCodeEntry as EdinetCodeEntry
from edinet.models.filing import Filing as Filing
from pydantic import BaseModel, computed_field

class Company(BaseModel):
    """提出者（企業）を表すモデル。"""
    model_config: Incomplete
    edinet_code: str
    name_ja: str | None
    sec_code: str | None
    @computed_field
    @property
    def ticker(self) -> str | None:
        """証券コード 4 桁を返す。"""
    @classmethod
    def from_filing(cls, filing: Filing) -> Company | None:
        """Filing から Company を構築する。

        Args:
            filing: 変換元の提出書類。

        Returns:
            `filing.edinet_code` がある場合は `Company`。無い場合は `None`。
        """
    def get_filings(self, date: str | DateType | None = None, *, start: str | DateType | None = None, end: str | DateType | None = None, doc_type: DocType | str | None = None) -> list[Filing]:
        """この企業の提出書類一覧を返す。

        Args:
            date: 単日指定。
            start: 範囲指定の開始日。
            end: 範囲指定の終了日。
            doc_type: 書類種別フィルタ。

        Returns:
            企業コードで絞り込まれた提出書類一覧。

        Note:
            `date` / `start` / `end` がすべて省略された場合は JST 当日を使う。
        """
    def latest(self, doc_type: DocType | str | None = None, *, start: str | DateType | None = None, end: str | DateType | None = None) -> Filing | None:
        '''この企業の最新の Filing を返す。

        ``get_filings()`` で書類一覧を取得し、``submit_date_time`` の降順で
        最初の Filing を返す。該当なしの場合は ``None``。

        ``start`` / ``end`` を省略した場合、過去 90 日間を検索する。

        Args:
            doc_type: 書類種別フィルタ。日本語文字列（``"有価証券報告書"``）
                または ``DocType`` Enum、コード文字列（``"120"``）を指定可能。
            start: 検索範囲の開始日。省略時は今日から 90 日前。
                ``end`` のみ指定した場合は ``end - 90日`` が自動設定される。
            end: 検索範囲の終了日。省略時は今日。
                ``start`` のみ指定した場合は今日が自動設定される。

        Returns:
            最新の Filing。見つからない場合は ``None``。

        Note:
            末尾（最新日）から1日ずつ探索し、該当書類が見つかった時点で返す。
            有報は決算期末の3ヶ月後（6月末）に集中するため、直近から探すと
            高速に見つかる。最悪ケースでも検索期間分の API コールが発生する。
        '''
    @classmethod
    def search(cls, query: str, *, limit: int = 20) -> list[Company]:
        '''企業名でCompanyを部分一致検索する。

        ``submitter_name`` / ``submitter_name_en`` / ``submitter_name_yomi`` の
        いずれかに ``query`` が含まれるエントリをCompanyに変換して返す。
        大文字・小文字を区別しない。

        結果は関連度順にソートされる:
        - 名前が query で始まるもの（prefix match）が先頭
        - 名前に query を含むもの（contains match）が後続

        Args:
            query: 検索キーワード。空文字の場合は空リストを返す。
            limit: 最大返却件数。0で無制限。デフォルト20。

        Returns:
            マッチしたCompanyのリスト（関連度順）。

        利用例:
            >>> results = Company.search("トヨタ")
            >>> results[0].name_ja
            \'トヨタ自動車株式会社\'
        '''
    @classmethod
    def from_edinet_code(cls, code: str) -> Company | None:
        '''EDINETコードからCompanyを構築する。

        Args:
            code: EDINETコード（例: ``"E02144"``）。

        Returns:
            Company。見つからない場合はNone。

        利用例:
            >>> company = Company.from_edinet_code("E02144")
            >>> company.name_ja
            \'トヨタ自動車株式会社\'
        '''
    @classmethod
    def from_sec_code(cls, sec_code: str) -> Company | None:
        '''証券コードからCompanyを構築する。

        4桁の場合は末尾に ``"0"`` を付与して5桁にしてから検索する。
        前後の空白は自動的に除去される。

        Args:
            sec_code: 証券コード（4桁または5桁）。

        Returns:
            Company。見つからない場合はNone。

        利用例:
            >>> company = Company.from_sec_code("7203")
            >>> company.edinet_code
            \'E02144\'
            >>> company = Company.from_sec_code("72030")  # 5桁でもOK
            >>> company.edinet_code
            \'E02144\'
        '''
    @classmethod
    def by_industry(cls, industry: str, *, limit: int = 100) -> list[Company]:
        '''業種名でCompanyを検索する。

        ``industry`` フィールドの完全一致で検索する。

        Args:
            industry: 業種名（例: ``"輸送用機器"``）。
            limit: 最大返却件数。0で無制限。デフォルト100。

        Returns:
            マッチしたCompanyのリスト。

        利用例:
            >>> results = Company.by_industry("輸送用機器")
            >>> len(results) > 0
            True
        '''
    @classmethod
    def all_listed(cls, *, limit: int = 0) -> list[Company]:
        """上場企業のCompany一覧を返す。

        Args:
            limit: 最大返却件数。0で無制限（デフォルト）。

        Returns:
            上場企業のCompanyのリスト。

        利用例:
            >>> listed = Company.all_listed()
            >>> len(listed) > 100
            True
        """



"""edinet パッケージの公開 API。"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from edinet._config import configure as configure
    from edinet._http import aclose as aclose
    from edinet.api.cache import CacheInfo as CacheInfo
    from edinet.api.cache import cache_info as cache_info
    from edinet.api.cache import clear_cache as clear_cache
    from edinet.models.company import Company as Company
    from edinet.models.doc_types import DocType as DocType
    from edinet.models.filing import Filing as Filing
    from edinet.models.financial import FinancialStatement as FinancialStatement
    from edinet.models.financial import LineItem as LineItem
    from edinet.models.revision import RevisionChain as RevisionChain
    from edinet.models.revision import build_revision_chain as build_revision_chain
    from edinet.public_api import adocuments as adocuments
    from edinet.public_api import documents as documents
    from edinet.financial.dimensions.fiscal_year import FiscalYearInfo as FiscalYearInfo
    from edinet.financial.dimensions.fiscal_year import (
        detect_fiscal_year as detect_fiscal_year,
    )
    from edinet.financial.statements import Statements as Statements
    from edinet.xbrl.taxonomy.custom import (
        CustomDetectionResult as CustomDetectionResult,
    )
    from edinet.xbrl.taxonomy.custom import (
        detect_custom_items as detect_custom_items,
    )
    from edinet.xbrl.taxonomy.custom import (
        find_custom_concepts as find_custom_concepts,
    )
    from edinet.xbrl.validation.calc_check import (
        CalcValidationResult as CalcValidationResult,
    )
    from edinet.xbrl.validation.calc_check import (
        validate_calculations as validate_calculations,
    )
    from edinet.financial.diff import DiffResult as DiffResult
    from edinet.financial.diff import diff_periods as diff_periods
    from edinet.financial.diff import diff_revisions as diff_revisions
    from edinet.financial.dimensions.segments import (
        extract_segments as extract_segments,
    )
    from edinet.financial.dimensions.segments import (
        list_dimension_axes as list_dimension_axes,
    )
    from edinet.financial.summary import FilingSummary as FilingSummary
    from edinet.financial.summary import build_summary as build_summary
    from edinet.exceptions import (
        EdinetError as EdinetError,
        EdinetConfigError as EdinetConfigError,
        EdinetAPIError as EdinetAPIError,
        EdinetParseError as EdinetParseError,
        EdinetWarning as EdinetWarning,
    )
    from edinet.financial.extract import (
        ExtractedValue as ExtractedValue,
        extract_values as extract_values,
        extracted_to_dict as extracted_to_dict,
    )
    from edinet.financial.standards.canonical_keys import CK as CK
    from edinet.financial.mapper import (
        summary_mapper as summary_mapper,
        statement_mapper as statement_mapper,
        definition_mapper as definition_mapper,
        calc_mapper as calc_mapper,
        dict_mapper as dict_mapper,
    )
    from edinet.xbrl.text import (
        extract_text_blocks as extract_text_blocks,
        build_section_map as build_section_map,
        clean_html as clean_html,
        TextBlock as TextBlock,
        SectionMap as SectionMap,
    )
    from edinet.financial.notes.employees import (
        extract_employee_info as extract_employee_info,
        EmployeeInfo as EmployeeInfo,
    )

_LAZY_EXPORTS: dict[str, tuple[str, str]] = {
    # --- 設定・ユーティリティ ---
    "configure": ("edinet._config", "configure"),
    "documents": ("edinet.public_api", "documents"),
    "adocuments": ("edinet.public_api", "adocuments"),
    "aclose": ("edinet._http", "aclose"),
    # --- コアモデル ---
    "Company": ("edinet.models.company", "Company"),
    "Filing": ("edinet.models.filing", "Filing"),
    "DocType": ("edinet.models.doc_types", "DocType"),
    "Statements": ("edinet.financial.statements", "Statements"),
    "FinancialStatement": ("edinet.models.financial", "FinancialStatement"),
    "LineItem": ("edinet.models.financial", "LineItem"),
    # --- Wave 6: キャッシュ (L2) ---
    "clear_cache": ("edinet.api.cache", "clear_cache"),
    "cache_info": ("edinet.api.cache", "cache_info"),
    "CacheInfo": ("edinet.api.cache", "CacheInfo"),
    # --- Wave 6: 訂正チェーン (L3) ---
    "build_revision_chain": ("edinet.models.revision", "build_revision_chain"),
    "RevisionChain": ("edinet.models.revision", "RevisionChain"),
    # --- Wave 6: 拡張科目検出 (L4) ---
    "detect_custom_items": ("edinet.xbrl.taxonomy.custom", "detect_custom_items"),
    "CustomDetectionResult": (
        "edinet.xbrl.taxonomy.custom",
        "CustomDetectionResult",
    ),
    # --- Wave 6: 計算バリデーション (L5) ---
    "validate_calculations": (
        "edinet.xbrl.validation.calc_check",
        "validate_calculations",
    ),
    "CalcValidationResult": (
        "edinet.xbrl.validation.calc_check",
        "CalcValidationResult",
    ),
    # --- Wave 6: 決算期判定 (L6) ---
    "detect_fiscal_year": (
        "edinet.financial.dimensions.fiscal_year",
        "detect_fiscal_year",
    ),
    "FiscalYearInfo": ("edinet.financial.dimensions.fiscal_year", "FiscalYearInfo"),
    # --- Wave 7: 拡張科目検出の拡張 (L5) ---
    "find_custom_concepts": (
        "edinet.xbrl.taxonomy.custom",
        "find_custom_concepts",
    ),
    # --- Wave 7: 差分比較 (L8) ---
    "diff_revisions": ("edinet.financial.diff", "diff_revisions"),
    "diff_periods": ("edinet.financial.diff", "diff_periods"),
    "DiffResult": ("edinet.financial.diff", "DiffResult"),
    # --- Wave 7: セグメント (L6) ---
    "extract_segments": (
        "edinet.financial.dimensions.segments",
        "extract_segments",
    ),
    "list_dimension_axes": (
        "edinet.financial.dimensions.segments",
        "list_dimension_axes",
    ),
    # --- Wave 7: サマリー (L4) ---
    "build_summary": ("edinet.financial.summary", "build_summary"),
    "FilingSummary": ("edinet.financial.summary", "FilingSummary"),
    # --- 例外 ---
    "EdinetError": ("edinet.exceptions", "EdinetError"),
    "EdinetConfigError": ("edinet.exceptions", "EdinetConfigError"),
    "EdinetAPIError": ("edinet.exceptions", "EdinetAPIError"),
    "EdinetParseError": ("edinet.exceptions", "EdinetParseError"),
    "EdinetWarning": ("edinet.exceptions", "EdinetWarning"),
    # --- 正規化キー抽出 ---
    "CK": ("edinet.financial.standards.canonical_keys", "CK"),
    "extract_values": ("edinet.financial.extract", "extract_values"),
    "extracted_to_dict": ("edinet.financial.extract", "extracted_to_dict"),
    "ExtractedValue": ("edinet.financial.extract", "ExtractedValue"),
    # --- マッパー ---
    "summary_mapper": ("edinet.financial.mapper", "summary_mapper"),
    "statement_mapper": ("edinet.financial.mapper", "statement_mapper"),
    "definition_mapper": ("edinet.financial.mapper", "definition_mapper"),
    "calc_mapper": ("edinet.financial.mapper", "calc_mapper"),
    "dict_mapper": ("edinet.financial.mapper", "dict_mapper"),
    # --- テキストブロック ---
    "extract_text_blocks": ("edinet.xbrl.text", "extract_text_blocks"),
    "build_section_map": ("edinet.xbrl.text", "build_section_map"),
    "clean_html": ("edinet.xbrl.text", "clean_html"),
    "TextBlock": ("edinet.xbrl.text", "TextBlock"),
    "SectionMap": ("edinet.xbrl.text", "SectionMap"),
    # --- 注記情報 ---
    "extract_employee_info": (
        "edinet.financial.notes.employees",
        "extract_employee_info",
    ),
    "EmployeeInfo": ("edinet.financial.notes.employees", "EmployeeInfo"),
}


def __getattr__(name: str):
    """公開シンボルを遅延ロードする。

    Args:
        name: 取得対象の属性名。

    Returns:
        遅延ロードされた属性。

    Raises:
        AttributeError: 公開対象外の属性名が指定された場合。
    """
    module_and_attr = _LAZY_EXPORTS.get(name)
    if module_and_attr is None:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    module_name, attr_name = module_and_attr
    module = __import__(module_name, fromlist=[attr_name])
    value = getattr(module, attr_name)
    globals()[name] = value
    return value


__all__ = [
    # 設定・ユーティリティ
    "configure",
    "documents",
    "adocuments",
    "aclose",
    # コアモデル
    "Company",
    "Filing",
    "DocType",
    "Statements",
    "FinancialStatement",
    "LineItem",
    # Wave 6: キャッシュ (L2)
    "clear_cache",
    "cache_info",
    "CacheInfo",
    # Wave 6: 訂正チェーン (L3)
    "build_revision_chain",
    "RevisionChain",
    # Wave 6: 拡張科目検出 (L4)
    "detect_custom_items",
    "CustomDetectionResult",
    # Wave 6: 計算バリデーション (L5)
    "validate_calculations",
    "CalcValidationResult",
    # Wave 6: 決算期判定 (L6)
    "detect_fiscal_year",
    "FiscalYearInfo",
    # Wave 7: 拡張科目検出の拡張 (L5)
    "find_custom_concepts",
    # Wave 7: 差分比較 (L8)
    "diff_revisions",
    "diff_periods",
    "DiffResult",
    # Wave 7: セグメント (L6)
    "extract_segments",
    "list_dimension_axes",
    # Wave 7: サマリー (L4)
    "build_summary",
    "FilingSummary",
    # 例外
    "EdinetError",
    "EdinetConfigError",
    "EdinetAPIError",
    "EdinetParseError",
    "EdinetWarning",
    # 正規化キー抽出
    "CK",
    "extract_values",
    "extracted_to_dict",
    "ExtractedValue",
    # マッパー
    "summary_mapper",
    "statement_mapper",
    "definition_mapper",
    "calc_mapper",
    "dict_mapper",
    # テキストブロック
    "extract_text_blocks",
    "build_section_map",
    "clean_html",
    "TextBlock",
    "SectionMap",
    # 注記情報
    "extract_employee_info",
    "EmployeeInfo",
]



"""EDINET ダウンロードファイルの永続ディスクキャッシュ。

``configure(cache_dir=...)`` で有効化すると、``Filing.fetch()`` が透過的に
ディスクキャッシュを使用する。キャッシュキーは ``doc_id``（書類管理番号）。

EDINET の開示書類は不変であり（訂正報告書は別の doc_id が付与される）、
TTL や LRU eviction は不要。

利用例::

    import edinet
    edinet.configure(cache_dir="~/.cache/edinet")

    filing.fetch()          # 1回目: ダウンロード → ディスク保存 → 返却
    filing.fetch()          # 2回目: ディスクから読み込み → 返却
    filing.fetch(refresh=True)  # 強制再ダウンロード → ディスク上書き

    from edinet.api.cache import clear_cache, cache_info
    info = cache_info()     # 統計情報
    clear_cache()           # downloads/ を全削除
"""

from __future__ import annotations

import logging
import os
import shutil
import tempfile
from dataclasses import dataclass
from pathlib import Path

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class CacheInfo:
    """キャッシュの統計情報。

    ``functools.lru_cache`` の ``cache_info()`` に倣ったパターン。

    Attributes:
        enabled: キャッシュが有効かどうか。
        cache_dir: キャッシュのルートディレクトリ。無効時は ``None``。
        entry_count: キャッシュエントリ数。
        total_bytes: キャッシュ合計バイト数。

    利用例::

        >>> from edinet.api.cache import cache_info
        >>> info = cache_info()
        >>> print(f"{info.entry_count} entries, {info.total_bytes / 1024 / 1024:.1f} MB")
    """

    enabled: bool
    cache_dir: Path | None
    entry_count: int
    total_bytes: int


class CacheStore:
    """ZIP ダウンロードのディスクキャッシュストア。

    キャッシュキーは ``doc_id``（書類管理番号）。
    EDINET の開示書類は不変であり（訂正報告書は別の doc_id が付与される）、
    TTL や LRU eviction は不要。

    原子的書き込み:
        ``put()`` は ``tempfile.NamedTemporaryFile`` で一時ファイルに書き込み、
        ``os.replace()`` でアトミックにリネームする。プロセスクラッシュ時に
        中途半端なファイルが残ることを防止する。

    Args:
        cache_dir: キャッシュのルートディレクトリ。

    利用例::

        >>> store = CacheStore(Path("/tmp/edinet_cache"))
        >>> store.put("S100ABC0", zip_bytes)
        >>> data = store.get("S100ABC0")
        >>> assert data == zip_bytes
    """

    def __init__(self, cache_dir: Path) -> None:
        self._cache_dir = cache_dir
        self._downloads_dir = cache_dir / "downloads"

    @property
    def cache_dir(self) -> Path:
        """キャッシュのルートディレクトリ。"""
        return self._cache_dir

    def cache_path(self, doc_id: str, *, suffix: str = ".zip") -> Path:
        """doc_id に対応するキャッシュファイルのパスを返す。

        Args:
            doc_id: 書類管理番号。
            suffix: ファイル拡張子。デフォルトは ``".zip"``。

        Returns:
            ``{cache_dir}/downloads/{doc_id}{suffix}``
        """
        return self._downloads_dir / f"{doc_id}{suffix}"

    def get(self, doc_id: str, *, suffix: str = ".zip") -> bytes | None:
        """キャッシュからバイナリを取得する。

        Args:
            doc_id: 書類管理番号。
            suffix: ファイル拡張子。デフォルトは ``".zip"``。

        Returns:
            キャッシュヒット時は bytes。ミス時は ``None``。
        """
        path = self.cache_path(doc_id, suffix=suffix)
        if not path.exists():
            return None
        try:
            data = path.read_bytes()
            logger.debug("キャッシュヒット: %s (%d bytes)", doc_id, len(data))
            return data
        except OSError:
            logger.warning("キャッシュ読み込み失敗: %s", doc_id, exc_info=True)
            return None

    def put(self, doc_id: str, data: bytes, *, suffix: str = ".zip") -> None:
        """バイナリをキャッシュに保存する。

        原子的書き込み: tempfile に書き込み後、``os.replace()`` でリネーム。

        Note:
            ``os.fsync()`` は意図的に省略。キャッシュは最適化であり耐久ストレージ
            ではないため、クラッシュ時にエントリが消失しても再ダウンロードすれば
            よい。fsync 省略により HDD 環境でのパフォーマンスを維持する。

        Args:
            doc_id: 書類管理番号。
            data: 保存するバイナリ。
            suffix: ファイル拡張子。デフォルトは ``".zip"``。
        """
        self._downloads_dir.mkdir(parents=True, exist_ok=True)
        target = self.cache_path(doc_id, suffix=suffix)
        tmp_name: str | None = None
        try:
            # 同一ディレクトリに一時ファイルを作成（os.replace のクロスデバイス問題を回避）
            fd = tempfile.NamedTemporaryFile(
                dir=self._downloads_dir,
                delete=False,
                suffix=".tmp",
            )
            tmp_name = fd.name
            try:
                fd.write(data)
                fd.flush()
            finally:
                fd.close()
            os.replace(tmp_name, target)
            tmp_name = None  # リネーム成功 → 掃除不要
            logger.debug("キャッシュ保存: %s (%d bytes)", doc_id, len(data))
        except OSError:
            # ディスク I/O エラー（容量不足、権限不足等）のみ吸収。
            # MemoryError 等の非 I/O エラーはここを通過して上位に伝播する。
            logger.warning("キャッシュ書き込み失敗: %s", doc_id, exc_info=True)
        finally:
            # 一時ファイルの掃除（リネーム前に例外が発生した場合）
            if tmp_name is not None:
                try:
                    os.unlink(tmp_name)
                except OSError:
                    pass

    def delete(self, doc_id: str, *, suffix: str = ".zip") -> None:
        """指定した doc_id のキャッシュを削除する。

        Args:
            doc_id: 書類管理番号。
            suffix: ファイル拡張子。デフォルトは ``".zip"``。
        """
        path = self.cache_path(doc_id, suffix=suffix)
        try:
            path.unlink(missing_ok=True)
            logger.debug("キャッシュ削除: %s", doc_id)
        except OSError:
            logger.warning("キャッシュ削除失敗: %s", doc_id, exc_info=True)

    def clear(self) -> None:
        """``downloads/`` サブディレクトリを削除する。

        Note:
            ``cache_dir`` 直下ではなく ``downloads/`` のみを削除する。
            ``TaxonomyResolver`` の pickle キャッシュ等、同じ ``cache_dir``
            を共有する他のキャッシュを巻き込まないための安全策。
        """
        if self._downloads_dir.exists():
            shutil.rmtree(self._downloads_dir)
            logger.info("キャッシュ全削除: %s", self._downloads_dir)

    def info(self) -> CacheInfo:
        """キャッシュの統計情報を返す。

        Returns:
            ``CacheInfo`` dataclass。
        """
        if not self._downloads_dir.exists():
            return CacheInfo(
                enabled=True,
                cache_dir=self._cache_dir,
                entry_count=0,
                total_bytes=0,
            )
        count = 0
        total = 0
        for f in self._downloads_dir.iterdir():
            if f.is_file() and not f.name.endswith(".tmp"):
                try:
                    total += f.stat().st_size
                    count += 1
                except OSError:
                    pass
        return CacheInfo(
            enabled=True,
            cache_dir=self._cache_dir,
            entry_count=count,
            total_bytes=total,
        )


def _get_cache_store() -> CacheStore | None:
    """キャッシュが有効なら ``CacheStore`` を返す。無効なら ``None``。

    ``Filing`` のヘルパーメソッドと ``clear_cache()`` / ``cache_info()`` が
    共通で使う内部関数。``get_config()`` + ``Path.expanduser()`` を 1 箇所に集約し、
    モック対象も 1 箇所にする。
    """
    from edinet._config import get_config

    config = get_config()
    if config.cache_dir is None:
        return None
    return CacheStore(Path(config.cache_dir).expanduser())


def clear_cache() -> None:
    """グローバル設定のダウンロードキャッシュを全削除する。

    ``configure(cache_dir=...)`` で設定されたキャッシュディレクトリ内の
    ``downloads/`` サブディレクトリを ``shutil.rmtree`` で削除する。
    ``cache_dir`` 直下の他のファイル（taxonomy pickle キャッシュ等）は
    保護される。キャッシュが無効の場合は何もしない。

    利用例::

        >>> import edinet
        >>> edinet.configure(cache_dir="/tmp/edinet_cache")
        >>> # ... 何らかの操作 ...
        >>> from edinet.api.cache import clear_cache
        >>> clear_cache()
    """
    store = _get_cache_store()
    if store is None:
        logger.debug("キャッシュ無効のため clear_cache() はスキップ")
        return
    store.clear()


def cache_info() -> CacheInfo:
    """キャッシュの統計情報を返す。

    ``functools.lru_cache`` の ``cache_info()`` に倣ったパターン。
    キャッシュ無効時は ``enabled=False`` で ``CacheInfo`` を返す。

    Returns:
        ``CacheInfo`` dataclass。

    利用例::

        >>> from edinet.api.cache import cache_info
        >>> info = cache_info()
        >>> info.entry_count
        42
    """
    store = _get_cache_store()
    if store is None:
        return CacheInfo(enabled=False, cache_dir=None, entry_count=0, total_bytes=0)
    return store.info()



"""内部入力バリデーションヘルパー。"""
from __future__ import annotations

import re

_EDINET_CODE_PATTERN = re.compile(r"^E\d{5}$")


def normalize_edinet_code(
    value: str | None,
    *,
    allow_none: bool = True,
) -> str | None:
    """edinet_code を正規化して検証する。

    Args:
        value: 入力値。
        allow_none: `None` を許可する場合は `True`。

    Returns:
        正規化済みの EDINET コード。`allow_none=True` かつ `value=None` の場合は `None`。

    Raises:
        ValueError: 型不正、空文字、または `E` + 5桁形式に一致しない場合。
    """
    if value is None:
        if allow_none:
            return None
        raise ValueError("edinet_code must not be None")

    if not isinstance(value, str):
        raise ValueError("edinet_code must be str or None")

    normalized = value.strip().upper()
    if not normalized:
        raise ValueError("edinet_code must not be empty")
    if not _EDINET_CODE_PATTERN.fullmatch(normalized):
        raise ValueError(
            f"Invalid edinet_code: {value!r}. Expected format like 'E02144'.",
        )
    return normalized



"""DEI ベースの期間分類モジュール。

DEI の日付情報から「当期」「前期」を分類し、
``Statements.income_statement(period="current")`` /
``period="prior"`` を可能にする。
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from edinet.xbrl.contexts import DurationPeriod, InstantPeriod

if TYPE_CHECKING:
    from edinet.xbrl.dei import DEI

__all__ = ["PeriodClassification", "classify_periods"]


@dataclass(frozen=True, slots=True)
class PeriodClassification:
    """DEI ベースの期間分類結果。

    Attributes:
        current_duration: 当期の duration 期間。
        prior_duration: 前期の duration 期間。
        current_instant: 当期末の instant 時点。
        prior_instant: 前期末の instant 時点。
    """

    current_duration: DurationPeriod | None
    prior_duration: DurationPeriod | None
    current_instant: InstantPeriod | None
    prior_instant: InstantPeriod | None


def classify_periods(dei: DEI) -> PeriodClassification:
    """DEI の日付情報から当期/前期を分類する。

    Args:
        dei: ``extract_dei()`` で取得した DEI。

    Returns:
        PeriodClassification。日付情報が不足している場合は
        該当フィールドが None になる。
    """
    # 当期 duration: start=current_fiscal_year_start_date, end=current_period_end_date
    current_duration = None
    if dei.current_fiscal_year_start_date and dei.current_period_end_date:
        current_duration = DurationPeriod(
            start_date=dei.current_fiscal_year_start_date,
            end_date=dei.current_period_end_date,
        )

    # 当期末 instant: instant=current_period_end_date
    current_instant = None
    if dei.current_period_end_date:
        current_instant = InstantPeriod(instant=dei.current_period_end_date)

    # 前期の終了日: comparative_period_end_date を優先、なければ previous_fiscal_year_end_date
    prior_end = dei.comparative_period_end_date or dei.previous_fiscal_year_end_date

    # 前期 duration: start=previous_fiscal_year_start_date, end=prior_end
    prior_duration = None
    if dei.previous_fiscal_year_start_date and prior_end:
        prior_duration = DurationPeriod(
            start_date=dei.previous_fiscal_year_start_date,
            end_date=prior_end,
        )

    # 前期末 instant: instant=prior_end
    prior_instant = None
    if prior_end:
        prior_instant = InstantPeriod(instant=prior_end)

    return PeriodClassification(
        current_duration=current_duration,
        prior_duration=prior_duration,
        current_instant=current_instant,
        prior_instant=prior_instant,
    )



"""決算期メタデータの抽出モジュール。

DEI の日付情報から決算月・期間月数・変則決算の有無を判定し、
``FiscalYearInfo`` として返す。

``classify_periods()`` がコンテキスト選択（BS/PL 構築時の当期/前期判別）
を担うのに対し、``detect_fiscal_year()`` は企業の決算期特性を表す
メタデータの抽出を担う。
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import date
from typing import TYPE_CHECKING, Any

from edinet.xbrl.dei import PeriodType

if TYPE_CHECKING:
    from edinet.xbrl.dei import DEI

__all__ = ["FiscalYearInfo", "detect_fiscal_year"]

# 平均月日数（365.25 / 12）。閏年を考慮した概算値。
_DAYS_PER_MONTH = 30.44


@dataclass(frozen=True, slots=True)
class FiscalYearInfo:
    """DEI から抽出した決算期メタデータ。

    DEI の日付情報を構造化し、決算期の特性（通期/半期、
    変則決算の有無、決算月等）を提供する。

    Attributes:
        start_date: 当事業年度開始日。DEI に未設定の場合は None。
        end_date: 当会計期間終了日。DEI に未設定の場合は None。
        fiscal_year_end_date: 決算期末日（当事業年度終了日）。
            DEI に未設定の場合は None。
        period_months: 期間月数（概算、四捨五入）。
            開始日または終了日が None の場合は None。
        period_type: 期間種類。DEI の type_of_current_period の値。
            PeriodType.FY（通期）、PeriodType.HY（半期）、
            未知の文字列、または None。
        is_full_year: 12ヶ月通期かどうか。
            PeriodType.FY かつ period_months が 12 の場合のみ True。
            日付情報が不足している場合は False。
        is_irregular: 変則決算かどうか。
            PeriodType.FY なのに period_months が 12 でない場合に True。
            日付情報が不足している場合は False。
        fiscal_year_end_month: 決算月（1〜12）。
            fiscal_year_end_date の月。未設定の場合は None。
    """

    start_date: date | None
    end_date: date | None
    fiscal_year_end_date: date | None
    period_months: int | None
    period_type: PeriodType | str | None
    is_full_year: bool
    is_irregular: bool
    fiscal_year_end_month: int | None


def _calc_months(start: date, end: date) -> int:
    """期間の月数を概算する。

    開始日・終了日の両端を含む日数を平均月日数（30.44）で
    割り、四捨五入して整数月数を返す。

    Args:
        start: 期間開始日。
        end: 期間終了日。

    Returns:
        概算月数（整数）。
    """
    delta_days = (end - start).days + 1  # 両端含む（会計期間の慣行）
    return round(delta_days / _DAYS_PER_MONTH)


def detect_fiscal_year(dei_or_stmts: DEI | Any) -> FiscalYearInfo:
    """DEI の日付情報から決算期メタデータを抽出する。

    ``Statements`` を渡すと内部の DEI を自動取得する。

    DEI の各日付フィールド（current_fiscal_year_start_date,
    current_period_end_date, current_fiscal_year_end_date,
    type_of_current_period）を構造化し、決算期の特性を判定する。

    日付フィールドが未設定（None）の場合は、該当する判定結果も
    None または False となる（graceful degradation）。

    Args:
        dei_or_stmts: ``extract_dei()`` で取得した DEI、
            または ``Statements`` オブジェクト。

    Returns:
        FiscalYearInfo。
    """
    # Statements を受け付ける
    from edinet.financial.statements import Statements as _Statements

    if isinstance(dei_or_stmts, _Statements):
        dei = dei_or_stmts.dei
        if dei is None:
            # DEI なしの Statements → 全フィールド None の FiscalYearInfo を返す
            return FiscalYearInfo(
                start_date=None, end_date=None,
                fiscal_year_end_date=None, period_months=None,
                period_type=None, is_full_year=False, is_irregular=False,
                fiscal_year_end_month=None,
            )
    else:
        dei = dei_or_stmts
    start = dei.current_fiscal_year_start_date
    end = dei.current_period_end_date
    fy_end = dei.current_fiscal_year_end_date
    period_type = dei.type_of_current_period

    # 期間月数の計算（両方の日付が必要）
    period_months: int | None = None
    if start is not None and end is not None:
        period_months = _calc_months(start, end)

    # 通期（FY）かどうかの判定
    is_fy = isinstance(period_type, PeriodType) and period_type == PeriodType.FY

    # 12ヶ月通期の判定
    is_full_year = is_fy and period_months == 12

    # 変則決算の判定（FY なのに 12ヶ月でない）
    is_irregular = is_fy and period_months is not None and period_months != 12

    # 決算月の抽出
    fiscal_year_end_month: int | None = None
    if fy_end is not None:
        fiscal_year_end_month = fy_end.month

    return FiscalYearInfo(
        start_date=start,
        end_date=end,
        fiscal_year_end_date=fy_end,
        period_months=period_months,
        period_type=period_type,
        is_full_year=is_full_year,
        is_irregular=is_irregular,
        fiscal_year_end_month=fiscal_year_end_month,
    )



from _typeshed import Incomplete

class EdinetWarning(UserWarning):
    '''edinet ライブラリが発行する warning の基底クラス。

    利用者が warnings.filterwarnings("ignore", category=EdinetWarning) で
    ライブラリ固有の warning だけをフィルタできる。
    '''
class EdinetError(Exception):
    """edinet ライブラリの基底例外。"""
class EdinetConfigError(EdinetError):
    """設定に関するエラー。"""

class EdinetAPIError(EdinetError):
    """EDINET API からのエラーレスポンス。"""
    status_code: Incomplete
    def __init__(self, status_code: int, message: str) -> None:
        """ステータスコードとメッセージを保持して初期化する。

        Args:
            status_code: HTTP ステータスコード。
            message: API が返したエラーメッセージ。
        """

class EdinetParseError(EdinetError):
    """取得済みデータ（JSON/ZIP/XBRL）の解析に失敗した。"""



from dataclasses import dataclass
from decimal import Decimal
from edinet.models.financial import FinancialStatement
from edinet.xbrl.linkbase.calculation import CalculationLinkbase
from typing import Literal

__all__ = ['CalcValidationResult', 'ValidationIssue', 'validate_calculations']

@dataclass(frozen=True, slots=True)
class ValidationIssue:
    '''計算バリデーションの個別不一致。

    Attributes:
        role_uri: 計算リンクベースのロール URI。
        parent_concept: 親概念のローカル名（例: ``"GrossProfit"``）。
        expected: 子科目の加重和（計算上の期待値）。
        actual: 実際の Fact 値。
        difference: ``|expected - actual|``。
        tolerance: decimals ベースの許容誤差。
        severity: 深刻度。tolerance 超過は ``"error"``。
        message: 人間向けメッセージ（日本語）。
    '''
    role_uri: str
    parent_concept: str
    expected: Decimal
    actual: Decimal
    difference: Decimal
    tolerance: Decimal
    severity: Literal['error', 'warning']
    message: str

@dataclass(frozen=True, slots=True)
class CalcValidationResult:
    """計算リンクベース検証の結果。

    Attributes:
        issues: 不一致の詳細タプル。
        checked_count: 検証した親科目の数（passed + error）。
        passed_count: 一致した（tolerance 内の）親科目の数。
        skipped_count: Fact 不足でスキップした親科目の数。
    """
    issues: tuple[ValidationIssue, ...]
    checked_count: int
    passed_count: int
    skipped_count: int
    @property
    def is_valid(self) -> bool:
        """error が 1 つもなければ True。"""
    @property
    def error_count(self) -> int:
        """error の件数。"""
    @property
    def warning_count(self) -> int:
        """warning の件数。"""

def validate_calculations(statement: FinancialStatement, calc_linkbase: CalculationLinkbase, *, role_uri: str | None = None) -> CalcValidationResult:
    """計算リンクベースに基づき財務諸表の合計値を検証する。

    XBRL 2.1 §5.2.5.2 に準拠し、計算リンクベース内の **全ての**
    summation-item 親子関係を検証対象とする（roots だけでなく
    中間ノードも含む）。各親科目について子科目の加重和と比較し、
    decimals ベースの丸め許容差を超える不一致を検出する。

    Args:
        statement: 検証対象の FinancialStatement。
        calc_linkbase: ``parse_calculation_linkbase()`` で取得した
            CalculationLinkbase。
        role_uri: 検証対象のロール URI。``None`` の場合は全ロールを検証する。

    Returns:
        CalcValidationResult。
    """



import enum
from dataclasses import dataclass
from edinet.xbrl._linkbase_utils import ROLE_LABEL as ROLE_LABEL, ROLE_TOTAL_LABEL, ROLE_VERBOSE_LABEL
from edinet.xbrl.taxonomy.concept_sets import ConceptEntry as ConceptEntry, ConceptSet as ConceptSet, ConceptSetRegistry as ConceptSetRegistry, StatementCategory as StatementCategory, classify_role_uri as classify_role_uri, derive_concept_sets as derive_concept_sets, derive_concept_sets_from_trees as derive_concept_sets_from_trees, get_concept_set as get_concept_set
from pathlib import Path

__all__ = ['LabelInfo', 'LabelSource', 'TaxonomyResolver', 'ROLE_LABEL', 'ROLE_VERBOSE', 'ROLE_TOTAL', 'ConceptEntry', 'ConceptSet', 'ConceptSetRegistry', 'StatementCategory', 'classify_role_uri', 'derive_concept_sets', 'derive_concept_sets_from_trees', 'get_concept_set']

type _LabelKey = tuple[str, str, str, str]
class LabelSource(enum.Enum):
    """ラベルの情報源。

    Attributes:
        STANDARD: EDINET 標準タクソノミ由来。
        FILER: 提出者別タクソノミ由来。
        FALLBACK: ラベルが見つからず local name を使用。
    """
    STANDARD = 'standard'
    FILER = 'filer'
    FALLBACK = 'fallback'

@dataclass(frozen=True, slots=True)
class LabelInfo:
    '''解決されたラベル情報。

    Attributes:
        text: ラベルテキスト（例: ``"売上高"``）。
        role: ラベルロール URI。
        lang: 言語コード（``"ja"`` / ``"en"``）。
        source: ラベルの情報源。
    '''
    text: str
    role: str
    lang: str
    source: LabelSource
ROLE_VERBOSE = ROLE_VERBOSE_LABEL
ROLE_TOTAL = ROLE_TOTAL_LABEL

class TaxonomyResolver:
    '''EDINET タクソノミのラベル解決を行うクラス。

    標準タクソノミの ``_lab.xml`` / ``_lab-en.xml`` をパースし、
    concept → ラベルの辞書を構築する。初回パース結果は pickle で
    キャッシュされ、2 回目以降は高速に読み込まれる。

    Attributes:
        taxonomy_version: タクソノミバージョン（例: ``"ALL_20251101"``）。
        taxonomy_path: タクソノミのルートパス。

    Example:
        >>> resolver = TaxonomyResolver("/path/to/ALL_20251101")
        >>> label = resolver.resolve("jppfs_cor", "NetSales")
        >>> print(label.text)
        売上高
    '''
    def __init__(self, taxonomy_path: str | Path, *, use_cache: bool = True) -> None:
        """TaxonomyResolver を初期化する。

        Args:
            taxonomy_path: タクソノミのルートディレクトリパス。
                ``ALL_20251101`` 等の最上位ディレクトリを指定する。
            use_cache: pickle キャッシュを使用するか。
                ``False`` の場合は毎回パースする（テスト用）。

        Raises:
            EdinetConfigError: taxonomy_path が存在しない場合。
        """
    @property
    def taxonomy_version(self) -> str:
        '''タクソノミバージョン文字列（例: ``"ALL_20251101"``）。'''
    @property
    def taxonomy_path(self) -> Path:
        """タクソノミのルートパス。"""
    def resolve(self, prefix: str, local_name: str, *, role: str = ..., lang: str = 'ja') -> LabelInfo:
        '''concept のラベルを解決する。

        Args:
            prefix: 名前空間プレフィックス（例: ``"jppfs_cor"``）。
            local_name: ローカル名（例: ``"NetSales"``）。
            role: ラベルロール URI。デフォルトは標準ラベル。
            lang: 言語コード。``"ja"`` または ``"en"``。

        Returns:
            解決された LabelInfo。ラベルが見つからない場合は
            指定 role → 標準ラベル → local name の順でフォールバック。

        Note:
            フォールバック時の LabelInfo は ``source=LabelSource.FALLBACK``、
            ``text=local_name`` となる。
        '''
    def resolve_clark(self, concept_qname: str, *, role: str = ..., lang: str = 'ja') -> LabelInfo:
        '''Clark notation の concept QName からラベルを解決する。

        Args:
            concept_qname: Clark notation の QName
                （例: ``"{http://...jppfs_cor}NetSales"``）。
            role: ラベルロール URI。
            lang: 言語コード。

        Returns:
            解決された LabelInfo。
        '''
    def load_filer_labels(self, lab_xml_bytes: bytes | None = None, lab_en_xml_bytes: bytes | None = None, *, xsd_bytes: bytes | None = None) -> int:
        """提出者別タクソノミのラベルを追加読み込みする。

        Args:
            lab_xml_bytes: 提出者の ``_lab.xml``（日本語）の bytes。
            lab_en_xml_bytes: 提出者の ``_lab-en.xml``（英語）の bytes。
            xsd_bytes: 提出者の ``.xsd`` の bytes。渡された場合、
                ``targetNamespace`` を抽出して ``_ns_to_prefix`` に
                追加する。

        Returns:
            追加されたラベル数。

        Warns:
            EdinetWarning: ``_filer_labels`` が空でない状態で呼ばれた場合。
        """
    def clear_filer_labels(self) -> None:
        """提出者別ラベルをクリアし、提出者由来の ``_ns_to_prefix`` エントリも除去する。

        次の filing を処理する前に呼び出す。
        """



"""XBRL 関連の名前空間定数と名前空間分類機能。

parser.py / contexts.py / taxonomy.py で共用する定数に加え、
名前空間 URI を解析して標準タクソノミ vs 提出者別タクソノミを判別する
分類機能を提供する。
"""

from __future__ import annotations

import enum
import re
from dataclasses import dataclass
from functools import lru_cache

# ============================================================
# 既存定数（変更厳禁 — parser.py / contexts.py / taxonomy で使用中）
# ============================================================

# XBRL Instance
NS_XBRLI = "http://www.xbrl.org/2003/instance"
# XBRL Dimensions
NS_XBRLDI = "http://xbrl.org/2006/xbrldi"
# XBRL Linkbase
NS_LINK = "http://www.xbrl.org/2003/linkbase"
# XLink
NS_XLINK = "http://www.w3.org/1999/xlink"
# XML Schema Instance
NS_XSI = "http://www.w3.org/2001/XMLSchema-instance"
# XML (xml:lang 等の属性用)
NS_XML = "http://www.w3.org/XML/1998/namespace"

# ============================================================
# 追加定数: XBRL 標準名前空間
# ============================================================

# ISO 4217 通貨コード
NS_ISO4217 = "http://www.xbrl.org/2003/iso4217"
# XBRL Dimensions Taxonomy
NS_XBRLDT = "http://xbrl.org/2005/xbrldt"
# XBRL 数値型（DTR: Data Type Registry）
NS_NUM = "http://www.xbrl.org/dtr/type/numeric"
# XBRL 非数値型（DTR: Data Type Registry）
NS_NONNUM = "http://www.xbrl.org/dtr/type/non-numeric"
# UTR (Unit Type Registry)
NS_UTR = "http://www.xbrl.org/2009/utr"

# ============================================================
# 追加定数: EDINET URI ベースパス
# ============================================================

EDINET_BASE = "http://disclosure.edinet-fsa.go.jp/"
"""EDINET の名前空間 URI のベースドメイン。

標準タクソノミ・提出者別タクソノミの両方がこのプレフィックスで始まる。
"""

EDINET_TAXONOMY_BASE = "http://disclosure.edinet-fsa.go.jp/taxonomy/"
"""EDINET 標準タクソノミの名前空間 URI はこのプレフィックスで始まる。"""

# ============================================================
# 内部定数
# ============================================================

_XBRL_INFRASTRUCTURE_URIS: frozenset[str] = frozenset({
    NS_XBRLI,
    NS_XBRLDI,
    NS_LINK,
    NS_XLINK,
    NS_XBRLDT,
})
"""XBRL インフラストラクチャに分類される名前空間 URI のセット。"""

# ============================================================
# Enum / Dataclass
# ============================================================


class NamespaceCategory(enum.Enum):
    """名前空間の分類カテゴリ。"""

    STANDARD_TAXONOMY = "standard_taxonomy"
    """EDINET 標準タクソノミ（jppfs_cor, jpcrp_cor, jpdei_cor, jpigp_cor 等）。"""

    FILER_TAXONOMY = "filer_taxonomy"
    """提出者別タクソノミ（企業固有の拡張科目）。"""

    XBRL_INFRASTRUCTURE = "xbrl_infrastructure"
    """XBRL 標準の基盤名前空間（xbrli, xlink, link, xbrldi 等）。"""

    OTHER = "other"
    """上記に該当しない名前空間（iso4217、XML Schema 等を含む）。"""


@dataclass(frozen=True, slots=True)
class NamespaceInfo:
    """名前空間 URI の解析結果。

    Attributes:
        uri: 元の名前空間 URI。
        category: 分類カテゴリ。
        is_standard: EDINET 標準タクソノミであれば True。
            STANDARD_TAXONOMY の場合のみ True。XBRL_INFRASTRUCTURE は False。
        module_name: タクソノミモジュール名（例: "jppfs_cor"）。
            STANDARD_TAXONOMY の場合のみ値を持つ。
        module_group: タクソノミモジュールグループ（例: "jppfs"）。
            STANDARD_TAXONOMY かつ正規表現マッチ時のみ値を持つ。
            後続レーン（standards/detect 等）での会計基準分岐に使用。
        taxonomy_version: タクソノミバージョン日付（例: "2025-11-01"）。
            STANDARD_TAXONOMY の場合のみ値を持つ。
        edinet_code: EDINET コード（例: "E02144"）。
            FILER_TAXONOMY の場合のみ値を持つ。
    """

    uri: str
    category: NamespaceCategory
    is_standard: bool
    module_name: str | None = None
    module_group: str | None = None
    taxonomy_version: str | None = None
    edinet_code: str | None = None


# ============================================================
# 正規表現パターン
# ============================================================

_STANDARD_TAXONOMY_PATTERN = re.compile(
    r"^http://disclosure\.edinet-fsa\.go\.jp/taxonomy/"
    r"(?P<module_group>[a-z]+)/"
    r"(?P<version>\d{4}-\d{2}-\d{2})/"
    r"(?P<module_name>[a-z_]+)$"
)
"""標準タクソノミ URI のパターン。

例: http://disclosure.edinet-fsa.go.jp/taxonomy/jppfs/2025-11-01/jppfs_cor
→ module_group="jppfs", version="2025-11-01", module_name="jppfs_cor"
"""

_FILER_TAXONOMY_PATTERN = re.compile(
    r"^http://disclosure\.edinet-fsa\.go\.jp/"
    r"(?!taxonomy/)"  # /taxonomy/ を含まない
    r"[^/]+/[^/]+/[^/]+/"
    r"(?P<edinet_code>[A-Z0-9]+)-(?P<branch>\d+)/"
)
"""提出者別タクソノミ URI のパターン。

例: http://disclosure.edinet-fsa.go.jp/jpcrp030000/asr/001/E02144-000/2025-03-31/01/2025-06-18
→ edinet_code="E02144", branch="000"
"""


# ============================================================
# 分類関数
# ============================================================


@lru_cache(maxsize=256)
def classify_namespace(uri: str) -> NamespaceInfo:
    """名前空間 URI を解析し、分類結果を返す。

    Args:
        uri: 名前空間 URI。

    Returns:
        解析結果の NamespaceInfo。
    """
    # 1. 標準タクソノミ（正規表現マッチ）
    m = _STANDARD_TAXONOMY_PATTERN.match(uri)
    if m:
        return NamespaceInfo(
            uri=uri,
            category=NamespaceCategory.STANDARD_TAXONOMY,
            is_standard=True,
            module_name=m.group("module_name"),
            module_group=m.group("module_group"),
            taxonomy_version=m.group("version"),
        )

    # 1b. 標準タクソノミ（フォールバック: EDINET_TAXONOMY_BASE で始まるが正規表現不一致）
    if uri.startswith(EDINET_TAXONOMY_BASE):
        return NamespaceInfo(
            uri=uri,
            category=NamespaceCategory.STANDARD_TAXONOMY,
            is_standard=True,
        )

    # 2. 提出者別タクソノミ（EDINET_BASE で始まり EDINET_TAXONOMY_BASE で始まらない）
    if uri.startswith(EDINET_BASE):
        m_filer = _FILER_TAXONOMY_PATTERN.match(uri)
        edinet_code = m_filer.group("edinet_code") if m_filer else None
        return NamespaceInfo(
            uri=uri,
            category=NamespaceCategory.FILER_TAXONOMY,
            is_standard=False,
            edinet_code=edinet_code,
        )

    # 3. XBRL インフラ名前空間
    if uri in _XBRL_INFRASTRUCTURE_URIS:
        return NamespaceInfo(
            uri=uri,
            category=NamespaceCategory.XBRL_INFRASTRUCTURE,
            is_standard=False,
        )

    # 4. その他
    return NamespaceInfo(
        uri=uri,
        category=NamespaceCategory.OTHER,
        is_standard=False,
    )


# ============================================================
# 便利関数（すべて classify_namespace に委譲しキャッシュの恩恵を受ける）
# ============================================================


def is_standard_taxonomy(uri: str) -> bool:
    """名前空間 URI が EDINET 標準タクソノミかどうかを返す。

    Args:
        uri: 名前空間 URI。

    Returns:
        標準タクソノミであれば True。
    """
    return classify_namespace(uri).is_standard


def is_filer_namespace(uri: str) -> bool:
    """名前空間 URI が提出者別タクソノミかどうかを返す。

    Args:
        uri: 名前空間 URI。

    Returns:
        提出者別タクソノミであれば True。
    """
    return classify_namespace(uri).category == NamespaceCategory.FILER_TAXONOMY


def extract_taxonomy_module(uri: str) -> str | None:
    """名前空間 URI からタクソノミモジュール名を抽出する。

    Args:
        uri: 名前空間 URI。

    Returns:
        モジュール名（例: "jppfs_cor"）。標準タクソノミでない場合 None。
    """
    return classify_namespace(uri).module_name


def extract_taxonomy_version(uri: str) -> str | None:
    """名前空間 URI からタクソノミバージョン日付を抽出する。

    Args:
        uri: 名前空間 URI。

    Returns:
        バージョン日付（例: "2025-11-01"）。標準タクソノミでない場合 None。
    """
    return classify_namespace(uri).taxonomy_version



"""財務諸表の HTML 表示モジュール。

Jupyter Notebook での ``_repr_html_()`` プロトコル用の HTML テーブル生成を提供する。
``display/statements.py`` の ``build_display_rows()`` を再利用し、
``DisplayRow`` → HTML ``<tr>`` への変換を行う。
"""

from __future__ import annotations

from decimal import Decimal
from html import escape
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from edinet.models.financial import FinancialStatement
    from edinet.xbrl.taxonomy.concept_sets import ConceptSet

__all__ = ["to_html"]

# インライン CSS（外部依存なし）
_CSS = """\
<style scoped>
.edinet-stmt { border-collapse: collapse; width: 100%; font-family: system-ui, sans-serif; font-size: 13px; }
.edinet-stmt th { background: #f5f5f5; text-align: left; padding: 6px 8px; border-bottom: 2px solid #ddd; }
.edinet-stmt td { padding: 4px 8px; border-bottom: 1px solid #eee; }
.edinet-stmt .amount { text-align: right; font-variant-numeric: tabular-nums; }
.edinet-stmt .abstract td { font-weight: bold; background: #fafafa; }
.edinet-stmt .total td { font-weight: bold; border-top: 1px solid #333; }
.edinet-stmt .label { white-space: nowrap; }
</style>
"""


def _format_value(value: Decimal | str | None) -> str:
    """値を表示用文字列に変換する。

    Args:
        value: 変換対象の値。Decimal は 3 桁カンマ区切り、
            str はエスケープ済み文字列、None は空文字列。

    Returns:
        表示用文字列。
    """
    if value is None:
        return ""
    if isinstance(value, Decimal):
        # 3桁カンマ区切り
        return f"{value:,.0f}"
    return escape(str(value))


def to_html(
    statement: FinancialStatement,
    *,
    concept_set: ConceptSet | None = None,
    abstract_labels: dict[str, str] | None = None,
) -> str:
    """FinancialStatement を HTML テーブル文字列に変換する。

    ``display/statements.py`` の ``build_display_rows()`` を使用して
    ``DisplayRow`` リストを生成し、HTML テーブルに変換する。

    Args:
        statement: 表示対象の財務諸表。
        concept_set: 階層表示用の ConceptSet。None の場合はフラット表示。
        abstract_labels: 抽象概念のラベル辞書。

    Returns:
        HTML テーブル文字列。
    """
    from edinet.display.statements import build_display_rows
    from edinet.models.financial import STATEMENT_TYPE_LABELS, format_period

    rows = build_display_rows(statement, concept_set, abstract_labels=abstract_labels)

    # タイトル構築
    type_label = STATEMENT_TYPE_LABELS.get(
        statement.statement_type,
        statement.statement_type.value,
    )
    scope = "連結" if statement.consolidated else "個別"
    period_str = format_period(statement.period) if statement.period else ""
    title = f"{type_label}  [{scope}] {period_str}"

    lines = [_CSS]
    lines.append('<table class="edinet-stmt">')
    lines.append(f"<caption><strong>{escape(title)}</strong></caption>")
    lines.append(
        "<thead><tr><th>科目</th><th class='amount'>金額</th></tr></thead>"
    )
    lines.append("<tbody>")

    for row in rows:
        css_classes: list[str] = []
        if row.is_abstract:
            css_classes.append("abstract")
        if row.is_total:
            css_classes.append("total")
        class_attr = f' class="{" ".join(css_classes)}"' if css_classes else ""

        indent = f"padding-left:{row.depth * 1.5}em" if row.depth > 0 else ""
        style_attr = f' style="{indent}"' if indent else ""

        value_str = _format_value(row.value)

        lines.append(f"<tr{class_attr}>")
        lines.append(f'  <td class="label"{style_attr}>{escape(row.label)}</td>')
        lines.append(f'  <td class="amount">{value_str}</td>')
        lines.append("</tr>")

    lines.append("</tbody></table>")
    return "\n".join(lines)



from dataclasses import dataclass
from edinet.xbrl.contexts import DurationPeriod, InstantPeriod
from edinet.xbrl.dei import DEI

__all__ = ['PeriodClassification', 'classify_periods']

@dataclass(frozen=True, slots=True)
class PeriodClassification:
    """DEI ベースの期間分類結果。

    Attributes:
        current_duration: 当期の duration 期間。
        prior_duration: 前期の duration 期間。
        current_instant: 当期末の instant 時点。
        prior_instant: 前期末の instant 時点。
    """
    current_duration: DurationPeriod | None
    prior_duration: DurationPeriod | None
    current_instant: InstantPeriod | None
    prior_instant: InstantPeriod | None

def classify_periods(dei: DEI) -> PeriodClassification:
    """DEI の日付情報から当期/前期を分類する。

    Args:
        dei: ``extract_dei()`` で取得した DEI。

    Returns:
        PeriodClassification。日付情報が不足している場合は
        該当フィールドが None になる。
    """



"""edinet.api の公開エントリ。"""

from edinet.api.cache import CacheInfo, CacheStore, cache_info, clear_cache
from edinet.api.documents import get_documents
from edinet.api.download import (
    DownloadFileType,
    download_document,
    extract_primary_xbrl,
    extract_zip_member,
    find_ixbrl_paths,
    find_primary_xbrl_path,
    list_zip_members,
)

__all__ = [
    "get_documents",
    "DownloadFileType",
    "download_document",
    "list_zip_members",
    "find_primary_xbrl_path",
    "find_ixbrl_paths",
    "extract_zip_member",
    "extract_primary_xbrl",
    # Wave 6: キャッシュ (L2)
    "CacheStore",
    "CacheInfo",
    "cache_info",
    "clear_cache",
]



"""DataFrame エクスポートユーティリティ。

CSV / Parquet / Excel への薄いラッパー。
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import pandas as pd

__all__ = ["to_csv", "to_parquet", "to_excel"]


def to_csv(
    df: pd.DataFrame,
    path: str | Path,
    *,
    encoding: str = "utf-8-sig",
    **kwargs: Any,
) -> None:
    """CSV 出力。

    デフォルトで ``utf-8-sig`` エンコーディングを使用し、
    Excel で開いた際の日本語文字化けを回避する。

    Args:
        df: 出力する DataFrame。
        path: 出力先ファイルパス。
        encoding: 文字エンコーディング。
        **kwargs: ``DataFrame.to_csv()`` に渡す追加引数。
    """
    df.to_csv(path, encoding=encoding, index=False, **kwargs)


def to_parquet(
    df: pd.DataFrame,
    path: str | Path,
    **kwargs: Any,
) -> None:
    """Parquet 出力。

    ``value`` 列に ``Decimal`` と ``str`` が混在する場合（Statements 全体の
    DataFrame 等）、pyarrow の型変換エラーを回避するため ``value`` 列を
    文字列に変換してから書き出す。

    Args:
        df: 出力する DataFrame。
        path: 出力先ファイルパス。
        **kwargs: ``DataFrame.to_parquet()`` に渡す追加引数。
    """
    out = df
    if "value" in df.columns and df["value"].apply(type).nunique() > 1:
        out = df.copy()
        out["value"] = out["value"].astype(str)
    out.to_parquet(path, index=False, **kwargs)


def to_excel(
    df: pd.DataFrame,
    path: str | Path,
    *,
    sheet_name: str = "Sheet1",
    **kwargs: Any,
) -> None:
    """Excel 出力。

    ``openpyxl`` が必要。インストールされていない場合は
    ``ImportError`` が発生する。

    Args:
        df: 出力する DataFrame。
        path: 出力先ファイルパス。
        sheet_name: シート名。
        **kwargs: ``DataFrame.to_excel()`` に渡す追加引数。
    """
    df.to_excel(path, sheet_name=sheet_name, index=False, **kwargs)



"""EDINET 提出書類 (Filing) の Pydantic モデル。

EDINET API 仕様書 §3-1-2-2 の全29フィールドを保持する。
"""
from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, ConfigDict, PrivateAttr, computed_field

from edinet.models.doc_types import DocType
from edinet.models.ordinance_code import OrdinanceCode

if TYPE_CHECKING:
    from edinet.models.company import Company
    from edinet.models.form_code import FormCodeEntry
    from edinet.financial.statements import Statements

# --- ヘルパー関数（モジュールプライベート） ---

def _parse_date(value: str | None) -> date | None:
    """EDINET API の日付文字列 'YYYY-MM-DD' を date に変換。

    None、空文字、空白のみの場合は None を返す。
    """
    if not value or not value.strip():
        return None
    return date.fromisoformat(value)

def _parse_datetime(value: str, *, field_name: str = "") -> datetime:
    """EDINET API の日時文字列 'YYYY-MM-DD HH:MM' を datetime に変換。

    Note:
        EDINET API は日本時間（JST, UTC+9）で日時を返すが、タイムゾーン情報は
        付与されない。本ライブラリでは naive datetime として保持する。
        期間比較や並び替えを行う場合は、全ての datetime が JST である
        前提で扱うこと。aware datetime への移行は v0.2.0 以降で検討。

    Raises:
        ValueError: フォーマット不正。field_name が指定されていれば
                    エラーメッセージにフィールド名と生文字列を含める。
    """
    try:
        return datetime.strptime(value, "%Y-%m-%d %H:%M")
    except ValueError as e:
        label = f" (field={field_name})" if field_name else ""
        msg = f"Invalid datetime format{label}: '{value}' (expected 'YYYY-MM-DD HH:MM')"
        raise ValueError(msg) from e
    
def _parse_datetime_optional(value: str | None, *, field_name: str = "") -> datetime | None:
    """EDINET API の日時文字列を datetime に変換。None/空なら None。"""
    if not value or not value.strip():
        return None
    return _parse_datetime(value, field_name=field_name)

def _parse_flag(value: str | None) -> bool:
    """EDINET API のフラグ '0'/'1' を bool に変換。"""
    return value == "1"

def _select_filer_xsd(candidates: list[str]) -> str | None:
    """複数の XSD 候補から提出者 XSD を決定的に選択する。

    EDINET コード（``_E\\d{5}``）を含むファイルを優先する。
    該当がなければアルファベット順で最初のファイルを返す。

    Args:
        candidates: PublicDoc 配下の非 audit XSD パスのリスト。

    Returns:
        選択されたパス。候補がない場合は None。
    """
    if not candidates:
        return None
    if len(candidates) == 1:
        return candidates[0]

    import re

    _EDINET_CODE_RE = re.compile(r"_E\d{5}", re.IGNORECASE)
    filer_matches = [c for c in candidates if _EDINET_CODE_RE.search(c)]
    if filer_matches:
        return sorted(filer_matches)[0]
    return sorted(candidates)[0]


def _extract_filer_taxonomy_files(zip_bytes: bytes | None) -> dict[str, bytes]:
    """ZIP 内から提出者別タクソノミファイルを抽出する。

    PublicDoc 配下の ``_lab.xml``（日本語ラベル）、``_lab-en.xml``（英語ラベル）、
    ``_cal.xml``（計算リンク）、``_def.xml``（定義リンク）、
    ``.xsd``（スキーマ、監査報告書を除く）を探索して返す。

    XSD が複数存在する場合は EDINET コードを含むファイルを優先し、
    それでも複数ある場合はアルファベット順で先頭のファイルを選択する。
    ラベルファイルは選択された XSD のベース名から導出し、XSD と同じ
    タクソノミに属するラベルを確実に取得する。

    Args:
        zip_bytes: ZIP のバイト列。None の場合は空辞書を返す。

    Returns:
        キーが ``"lab"`` / ``"lab_en"`` / ``"xsd"`` / ``"cal"`` / ``"def"``、値が bytes の辞書。
        対応するファイルが存在しない場合はキーが含まれない。
    """
    if zip_bytes is None:
        return {}

    import logging

    from edinet.api.download import extract_zip_member, list_zip_members

    logger = logging.getLogger(__name__)
    members = list_zip_members(zip_bytes)
    member_set = set(members)
    xsd_candidates: list[str] = []
    lab_fallbacks: list[str] = []
    lab_en_fallbacks: list[str] = []

    for name in members:
        lower = name.lower()
        if "publicdoc/" not in lower.replace("\\", "/"):
            continue
        if lower.endswith("_lab.xml") and "_lab-en" not in lower:
            lab_fallbacks.append(name)
        elif lower.endswith("_lab-en.xml"):
            lab_en_fallbacks.append(name)
        elif lower.endswith(".xsd") and "audit" not in lower:
            xsd_candidates.append(name)

    result: dict[str, bytes] = {}

    # 1. XSD を決定的に選択
    selected_xsd = _select_filer_xsd(xsd_candidates)
    if selected_xsd is not None:
        result["xsd"] = extract_zip_member(zip_bytes, selected_xsd)

    if len(xsd_candidates) > 1:
        logger.debug(
            "複数の XSD を検出: %s → %s を選択", xsd_candidates, selected_xsd,
        )

    # 2. ラベル・リンクベースファイルを XSD のベース名から導出
    if selected_xsd is not None:
        xsd_stem = selected_xsd[:selected_xsd.lower().rfind(".xsd")]
        derived_lab = xsd_stem + "_lab.xml"
        derived_lab_en = xsd_stem + "_lab-en.xml"
        derived_cal = xsd_stem + "_cal.xml"
        derived_def = xsd_stem + "_def.xml"

        if derived_lab in member_set:
            result["lab"] = extract_zip_member(zip_bytes, derived_lab)
        if derived_lab_en in member_set:
            result["lab_en"] = extract_zip_member(zip_bytes, derived_lab_en)
        if derived_cal in member_set:
            result["cal"] = extract_zip_member(zip_bytes, derived_cal)
        if derived_def in member_set:
            result["def"] = extract_zip_member(zip_bytes, derived_def)

    # 3. XSD から導出できなかった場合はフォールバック
    if "lab" not in result and lab_fallbacks:
        result["lab"] = extract_zip_member(zip_bytes, sorted(lab_fallbacks)[0])
    if "lab_en" not in result and lab_en_fallbacks:
        result["lab_en"] = extract_zip_member(zip_bytes, sorted(lab_en_fallbacks)[0])

    logger.debug("filer taxonomy files found: %s", list(result.keys()))
    return result


class Filing(BaseModel):
    """EDINET に提出された書類1件を表すモデル。

    EDINET API の書類一覧レスポンス (results 配列の1要素) から生成する。
    API 由来の全29フィールドを保持し、データを一切落とさない。
    加えて派生フィールド（doc_type, filing_date, ticker, doc_type_label_ja）を
    computed_field として提供する。
    また、`ordinance` / `form` は model_dump に含めない `@property` として提供する。

    **Note: コード値の型について**
    - `docTypeCode`: `DocType` Enum に変換して保持する（Day 4 のスコープ）。
    - `ordinanceCode`, `formCode`, `fundCode`: Day 4 時点では `str` としてそのまま保持する。これらは定義数が多く更新頻度も高いため、将来的にメタプログラミングで生成された専用 Enum/Model に移行する予定（Day 4.5 以降）。


    Note:
        日時フィールド（submit_date_time, ope_date_time）は EDINET API が返す
        JST（日本時間, UTC+9）を naive datetime として保持する。
        タイムゾーン情報は付与されない。

    利用例:
        >>> result = get_documents("2026-02-07")
        >>> filings = Filing.from_api_list(result)
        >>> f = filings[0]
        >>> f.doc_id
        'S100XXXX'
        >>> f.doc_type
        DocType.ANNUAL_SECURITIES_REPORT
        >>> f.filing_date
        datetime.date(2026, 2, 7)
    """

    model_config = ConfigDict(frozen=True)

    # --- 識別子 ---
    seq_number: int
    doc_id: str

    # --- 書類種別 ---
    doc_type_code: str | None
    ordinance_code: str | None
    form_code: str | None

    # --- 提出者情報（Day 6 で Company にネストする候補） ---
    edinet_code: str | None
    sec_code: str | None
    jcn: str | None
    filer_name: str | None
    fund_code: str | None

    # --- 日付・期間 ---
    submit_date_time: datetime
    period_start: date | None
    period_end: date | None

    # --- 書類概要 ---
    doc_description: str | None

    # --- 関連書類・関連コード ---
    issuer_edinet_code: str | None
    subject_edinet_code: str | None
    subsidiary_edinet_code: str | None
    current_report_reason: str | None
    parent_doc_id: str | None
    ope_date_time: datetime | None

    # --- ステータス（Day 4 では文字列のまま） ---
    withdrawal_status: str
    doc_info_edit_status: str
    disclosure_status: str

    # --- コンテンツフラグ ---
    has_xbrl: bool
    has_pdf: bool
    has_attachment: bool
    has_english: bool
    has_csv: bool

    # --- 法定・任意 ---
    legal_status: str

    # --- Day 6: fetch キャッシュ ---
    _zip_cache: bytes | None = PrivateAttr(default=None)
    _xbrl_cache: tuple[str, bytes] | None = PrivateAttr(default=None)
    _pdf_cache: bytes | None = PrivateAttr(default=None)

    # --- 計算フィールド ---

    @computed_field  # type: ignore[prop-decorator]
    @property
    def doc_type(self) -> DocType | None:
        """doc_type_code に対応する DocType Enum。

        未知のコードの場合は None（DocType.from_code() が warning を出す）。
        コード自体が None の場合も None。
        dict lookup（O(1)）なのでキャッシュ不要。
        """
        if self.doc_type_code is None:
            return None
        return DocType.from_code(self.doc_type_code)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def filing_date(self) -> date:
        """提出日（submit_date_time の日付部分）。

        PLAN.md §4 の Filing モデル定義で filing_date: date と定義。
        submit_date_time から導出することで二重管理を避ける。
        """
        return self.submit_date_time.date()

    @property
    def ordinance(self) -> OrdinanceCode | None:
        """府令コード Enum。"""
        if self.ordinance_code is None:
            return None
        return OrdinanceCode.from_code(self.ordinance_code)

    @property
    def form(self) -> FormCodeEntry | None:
        """様式コード情報。"""
        if self.ordinance_code is None or self.form_code is None:
            return None
        from edinet.models.form_code import get_form_code

        return get_form_code(self.ordinance_code, self.form_code)

    @property
    def company(self) -> Company | None:
        """提出者 Company を返す。"""
        from edinet.models.company import Company
        return Company.from_filing(self)

    def clear_fetch_cache(self) -> None:
        """`fetch()` / `fetch_pdf()` のキャッシュを破棄する。"""
        object.__setattr__(self, "_zip_cache", None)
        object.__setattr__(self, "_xbrl_cache", None)
        object.__setattr__(self, "_pdf_cache", None)

    def _get_from_disk_cache(self) -> bytes | None:
        """ディスクキャッシュから ZIP を読み込む。

        Returns:
            キャッシュヒット時は bytes。キャッシュ無効またはミス時は ``None``。
        """
        from edinet.api.cache import _get_cache_store

        store = _get_cache_store()
        if store is None:
            return None
        return store.get(self.doc_id)

    def _save_to_disk_cache(self, data: bytes) -> None:
        """ZIP をディスクキャッシュに保存する。

        Args:
            data: 保存する ZIP バイナリ。
        """
        from edinet.api.cache import _get_cache_store

        store = _get_cache_store()
        if store is None:
            return
        store.put(self.doc_id, data)

    def _delete_disk_cache(self) -> None:
        """ディスクキャッシュの当該エントリを削除する。

        破損したキャッシュファイルによる無限ループを防止するために使用。
        ディスクキャッシュ由来の ZIP が解析に失敗した場合に呼ばれ、
        次回の ``fetch()`` でネットワークから再取得できるようにする。
        """
        from edinet.api.cache import _get_cache_store

        store = _get_cache_store()
        if store is None:
            return
        store.delete(self.doc_id)

    def _get_pdf_from_disk_cache(self) -> bytes | None:
        """ディスクキャッシュから PDF を読み込む。

        Returns:
            キャッシュヒット時は bytes。キャッシュ無効またはミス時は ``None``。
        """
        from edinet.api.cache import _get_cache_store

        store = _get_cache_store()
        if store is None:
            return None
        return store.get(self.doc_id, suffix=".pdf")

    def _save_pdf_to_disk_cache(self, data: bytes) -> None:
        """PDF をディスクキャッシュに保存する。

        Args:
            data: 保存する PDF バイナリ。
        """
        from edinet.api.cache import _get_cache_store

        store = _get_cache_store()
        if store is None:
            return
        store.put(self.doc_id, data, suffix=".pdf")

    def _delete_pdf_disk_cache(self) -> None:
        """ディスクキャッシュの PDF エントリを削除する。"""
        from edinet.api.cache import _get_cache_store

        store = _get_cache_store()
        if store is None:
            return
        store.delete(self.doc_id, suffix=".pdf")

    def fetch_pdf(self, *, refresh: bool = False) -> bytes:
        """提出書類の PDF をダウンロードして返す。

        3層キャッシュ（メモリ → ディスク → ネットワーク）を使用する。

        Args:
            refresh: ``True`` の場合は PDF キャッシュを破棄して再取得する。

        Returns:
            PDF のバイト列。

        Raises:
            EdinetAPIError: 当該書類に PDF が含まれていない場合。
            EdinetParseError: ダウンロードの入力不正の正規化。
            EdinetError: 通信層失敗。
        """
        from edinet.api.download import DownloadFileType, download_document
        from edinet.exceptions import EdinetAPIError, EdinetParseError

        if not self.has_pdf:
            raise EdinetAPIError(
                0,
                f"書類に PDF が含まれていません (doc_id={self.doc_id}, pdfFlag=0)。",
            )

        if refresh:
            object.__setattr__(self, "_pdf_cache", None)
            self._delete_pdf_disk_cache()

        # Layer 1: メモリキャッシュ
        if self._pdf_cache is not None:
            return self._pdf_cache

        # Layer 2: ディスクキャッシュ
        if not refresh:
            disk_data = self._get_pdf_from_disk_cache()
            if disk_data is not None:
                object.__setattr__(self, "_pdf_cache", disk_data)
                return disk_data

        # Layer 3: ネットワーク
        try:
            pdf_bytes = download_document(
                self.doc_id,
                file_type=DownloadFileType.PDF,
            )
        except ValueError as exc:
            raise EdinetParseError(
                f"PDF のダウンロードに失敗しました "
                f"(doc_id={self.doc_id!r}): {exc}",
            ) from exc

        object.__setattr__(self, "_pdf_cache", pdf_bytes)
        self._save_pdf_to_disk_cache(pdf_bytes)
        return pdf_bytes

    async def afetch_pdf(self, *, refresh: bool = False) -> bytes:
        """提出書類の PDF を非同期でダウンロードして返す。

        ``fetch_pdf()`` の非同期版。

        Args:
            refresh: ``True`` の場合は PDF キャッシュを破棄して再取得する。

        Returns:
            PDF のバイト列。

        Raises:
            EdinetAPIError: 当該書類に PDF が含まれていない場合。
            EdinetParseError: ダウンロードの入力不正の正規化。
            EdinetError: 通信層失敗。
        """
        from edinet.api.download import DownloadFileType, adownload_document
        from edinet.exceptions import EdinetAPIError, EdinetParseError

        if not self.has_pdf:
            raise EdinetAPIError(
                0,
                f"書類に PDF が含まれていません (doc_id={self.doc_id}, pdfFlag=0)。",
            )

        if refresh:
            object.__setattr__(self, "_pdf_cache", None)
            self._delete_pdf_disk_cache()

        # Layer 1: メモリキャッシュ
        if self._pdf_cache is not None:
            return self._pdf_cache

        # Layer 2: ディスクキャッシュ
        if not refresh:
            disk_data = self._get_pdf_from_disk_cache()
            if disk_data is not None:
                object.__setattr__(self, "_pdf_cache", disk_data)
                return disk_data

        # Layer 3: ネットワーク
        try:
            pdf_bytes = await adownload_document(
                self.doc_id,
                file_type=DownloadFileType.PDF,
            )
        except ValueError as exc:
            raise EdinetParseError(
                f"PDF のダウンロードに失敗しました "
                f"(doc_id={self.doc_id!r}): {exc}",
            ) from exc

        object.__setattr__(self, "_pdf_cache", pdf_bytes)
        self._save_pdf_to_disk_cache(pdf_bytes)
        return pdf_bytes

    def fetch(self, *, refresh: bool = False) -> tuple[str, bytes]:
        """提出本文書 ZIP から代表 XBRL を取得する。

        Args:
            refresh: `True` の場合はキャッシュを破棄して再取得する。

        Returns:
            `(xbrl_path, xbrl_bytes)` のタプル。

        Raises:
            EdinetAPIError: 当該書類に XBRL が無い場合。
            EdinetError: 通信層失敗。
            EdinetAPIError: API レイヤー失敗。
            EdinetParseError: ZIP 解析失敗、primary 不在、または入力不正の正規化。
        """
        from edinet.api.download import (
            DownloadFileType,
            download_document,
            extract_primary_xbrl,
        )
        from edinet.exceptions import EdinetAPIError, EdinetParseError

        if not self.has_xbrl:
            raise EdinetAPIError(
                0,
                f"書類に XBRL が含まれていません (doc_id={self.doc_id}, xbrlFlag=0)。",
            )

        if refresh:
            self.clear_fetch_cache()

        if self._xbrl_cache is not None:
            return self._xbrl_cache

        from_disk = False

        if self._zip_cache is None:
            # Layer 2: ディスクキャッシュ確認
            if not refresh:
                disk_data = self._get_from_disk_cache()
                if disk_data is not None:
                    object.__setattr__(self, "_zip_cache", disk_data)
                    from_disk = True

        if self._zip_cache is None:
            try:
                zip_bytes = download_document(
                    self.doc_id,
                    file_type=DownloadFileType.XBRL_AND_AUDIT,
                )
            except ValueError as exc:
                raise EdinetParseError(
                    f"XBRL ZIP のダウンロードに失敗しました "
                    f"(doc_id={self.doc_id!r}): {exc}",
                ) from exc
            object.__setattr__(self, "_zip_cache", zip_bytes)
            # Layer 2: ディスクに保存
            self._save_to_disk_cache(zip_bytes)

        assert self._zip_cache is not None
        try:
            result = extract_primary_xbrl(self._zip_cache)
        except ValueError as exc:
            if from_disk:
                self._delete_disk_cache()
            self.clear_fetch_cache()
            raise EdinetParseError(
                f"EDINET ZIP の解析に失敗しました (doc_id={self.doc_id})。",
            ) from exc

        if result is None:
            if from_disk:
                self._delete_disk_cache()
            self.clear_fetch_cache()
            raise EdinetParseError(
                f"ZIP 内に主要な XBRL が見つかりません (doc_id={self.doc_id})。",
            )

        object.__setattr__(self, "_xbrl_cache", result)
        return result

    async def afetch(self, *, refresh: bool = False) -> tuple[str, bytes]:
        """提出本文書 ZIP から代表 XBRL を非同期で取得する。

        Args:
            refresh: `True` の場合はキャッシュを破棄して再取得する。

        Returns:
            `(xbrl_path, xbrl_bytes)` のタプル。

        Raises:
            EdinetAPIError: 当該書類に XBRL が無い場合。
            EdinetError: 通信層失敗。
            EdinetAPIError: API レイヤー失敗。
            EdinetParseError: ZIP 解析失敗、primary 不在、または入力不正の正規化。
        """
        from edinet.api.download import (
            DownloadFileType,
            adownload_document,
            extract_primary_xbrl,
        )
        from edinet.exceptions import EdinetAPIError, EdinetParseError

        if not self.has_xbrl:
            raise EdinetAPIError(
                0,
                f"書類に XBRL が含まれていません (doc_id={self.doc_id}, xbrlFlag=0)。",
            )

        if refresh:
            self.clear_fetch_cache()

        if self._xbrl_cache is not None:
            return self._xbrl_cache

        from_disk = False

        if self._zip_cache is None:
            # Layer 2: ディスクキャッシュ確認（同期 I/O で十分）
            if not refresh:
                disk_data = self._get_from_disk_cache()
                if disk_data is not None:
                    object.__setattr__(self, "_zip_cache", disk_data)
                    from_disk = True

        if self._zip_cache is None:
            try:
                zip_bytes = await adownload_document(
                    self.doc_id,
                    file_type=DownloadFileType.XBRL_AND_AUDIT,
                )
            except ValueError as exc:
                raise EdinetParseError(
                    f"XBRL ZIP のダウンロードに失敗しました "
                    f"(doc_id={self.doc_id!r}): {exc}",
                ) from exc
            object.__setattr__(self, "_zip_cache", zip_bytes)
            # Layer 2: ディスクに保存（同期 I/O で十分）
            self._save_to_disk_cache(zip_bytes)

        assert self._zip_cache is not None
        try:
            result = extract_primary_xbrl(self._zip_cache)
        except ValueError as exc:
            if from_disk:
                self._delete_disk_cache()
            self.clear_fetch_cache()
            raise EdinetParseError(
                f"EDINET ZIP の解析に失敗しました (doc_id={self.doc_id})。",
            ) from exc

        if result is None:
            if from_disk:
                self._delete_disk_cache()
            self.clear_fetch_cache()
            raise EdinetParseError(
                f"ZIP 内に主要な XBRL が見つかりません (doc_id={self.doc_id})。",
            )

        object.__setattr__(self, "_xbrl_cache", result)
        return result

    @computed_field  # type: ignore[prop-decorator]
    @property
    def ticker(self) -> str | None:
        """証券コード4桁。

        sec_code は5桁（末尾チェックディジット付き）で返る。
        先頭4桁を切り出して一般的な「証券コード」として返す。
        sec_code が None または短すぎる場合は None。
        """
        if self.sec_code and len(self.sec_code) >= 4:
            return self.sec_code[:4]
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def doc_type_label_ja(self) -> str:
        """書類種別の日本語表示ラベル（常に str を返す）。

        doc_type が None（未知コードまたはコードなし）の場合は
        doc_type_code にフォールバックするが、それも None なら空文字を返す。
        UI やログで doc_type の Optional チェックを毎回書かずに済む。
        """
        dt = self.doc_type
        if dt:
            return dt.name_ja
        return self.doc_type_code or ""
    
    # --- XBRL パイプライン ---

    def _resolve_taxonomy_path(self, taxonomy_path: str | None) -> str:
        """taxonomy_path を解決する。引数 > configure() の優先順位。

        Args:
            taxonomy_path: 明示指定されたパス。None の場合は configure() から取得。

        Returns:
            解決されたタクソノミパス。

        Raises:
            EdinetConfigError: どちらも未設定の場合。
        """
        from edinet._config import get_config
        from edinet.exceptions import EdinetConfigError

        resolved = taxonomy_path or get_config().taxonomy_path
        if resolved is None:
            raise EdinetConfigError(
                "taxonomy_path が未設定です。"
                "xbrl(taxonomy_path=...) で直接指定するか、"
                "configure(taxonomy_path=...) でグローバルに設定してください。"
            )
        return resolved

    def _build_statements(
        self,
        taxonomy_path: str,
        xbrl_path: str,
        xbrl_bytes: bytes,
    ) -> Statements:
        """ZIP キャッシュからパイプラインを実行する（同期）。

        xbrl() / axbrl() の共通パイプライン。I/O 完了後の
        ステップ 3〜8 を担当する。

        Args:
            taxonomy_path: 解決済みタクソノミパス。
            xbrl_path: XBRL ファイルのパス（トレース用）。
            xbrl_bytes: XBRL ファイルのバイト列。

        Returns:
            Statements コンテナ。

        Raises:
            EdinetParseError: パイプラインの各ステップで失敗した場合。
        """
        import logging

        from edinet.exceptions import EdinetError, EdinetParseError
        from edinet.xbrl import (
            TaxonomyResolver,
            build_line_items,
            build_statements,
            parse_xbrl_facts,
            structure_contexts,
        )

        logger = logging.getLogger(__name__)

        try:
            # 3. 提出者タクソノミファイルの抽出
            filer_files = _extract_filer_taxonomy_files(self._zip_cache)

            # 4. XBRL パース
            parsed = parse_xbrl_facts(xbrl_bytes, source_path=xbrl_path)
            logger.debug(
                "step 4: parsed %d facts, %d contexts",
                len(parsed.facts), len(parsed.contexts),
            )

            # 4b. J-GAAP 名前空間チェック（IFRS/US-GAAP 検知）
            namespaces = {
                f.concept_qname.split("}")[0]
                for f in parsed.facts
                if "}" in f.concept_qname
            }
            if not any("jppfs" in ns for ns in namespaces):
                import warnings

                from edinet.exceptions import EdinetWarning

                warnings.warn(
                    f"Filing {self.doc_id}: jppfs_cor 名前空間の Fact がありません。"
                    "IFRS / US-GAAP の Filing は v0.1.0 では未対応です。",
                    EdinetWarning,
                    stacklevel=3,
                )

            # 5. Context 構造化
            ctx_map = structure_contexts(parsed.contexts)
            logger.debug("step 5: structured %d contexts", len(ctx_map))

            # 6. ラベル解決
            resolver = TaxonomyResolver(taxonomy_path)
            resolver.load_filer_labels(
                lab_xml_bytes=filer_files.get("lab"),
                lab_en_xml_bytes=filer_files.get("lab_en"),
                xsd_bytes=filer_files.get("xsd"),
            )

            # 7. LineItem 生成
            items = build_line_items(parsed.facts, ctx_map, resolver)
            logger.debug("step 7: built %d line items", len(items))

            # 7b. リンクベースパース（definition_mapper / calc_mapper 用）
            from edinet.xbrl.linkbase import (
                parse_calculation_linkbase,
                parse_definition_linkbase,
            )

            cal_lb = None
            cal_bytes = filer_files.get("cal")
            if cal_bytes is not None:
                cal_lb = parse_calculation_linkbase(cal_bytes)

            def_trees = None
            def_bytes = filer_files.get("def")
            if def_bytes is not None:
                def_trees = parse_definition_linkbase(def_bytes)

            logger.debug(
                "step 7b: linkbases cal=%s def=%s",
                cal_lb is not None,
                def_trees is not None,
            )

            # 8. Statement 組み立て
            from pathlib import Path as _Path

            from edinet.xbrl.dei import extract_dei, resolve_industry_code

            dei = extract_dei(parsed.facts)
            industry_code = resolve_industry_code(dei)

            stmts = build_statements(
                items,
                facts=parsed.facts,
                contexts=ctx_map,
                taxonomy_root=_Path(taxonomy_path),
                industry_code=industry_code,
                resolver=resolver,
                calculation_linkbase=cal_lb,
                definition_linkbase=def_trees,
            )
        except EdinetError:
            raise
        except Exception as exc:
            raise EdinetParseError(
                f"Filing {self.doc_id}: XBRL パイプラインで予期しないエラーが発生しました"
            ) from exc

        return stmts

    def xbrl(self, *, taxonomy_path: str | None = None) -> Statements:
        """XBRL を解析し財務諸表コンテナを返す。

        ZIP ダウンロード → XBRL パース → Context 構造化 → ラベル解決 →
        LineItem 生成 → Statement 組み立てのフルパイプラインを実行する。

        Args:
            taxonomy_path: EDINET タクソノミのルートパス
                （例: ``"/path/to/ALL_20251101"``）。
                省略時は ``configure(taxonomy_path=...)`` で設定された値を使用する。

        Returns:
            Statements コンテナ。``income_statement()`` / ``balance_sheet()`` /
            ``cash_flow_statement()`` でアクセスする。

        Raises:
            EdinetAPIError: 当該書類に XBRL が無い場合。
            EdinetConfigError: taxonomy_path が未設定の場合。
            EdinetParseError: XBRL パースに失敗した場合。
            EdinetError: 通信層の失敗。

        Note:
            J-GAAP / IFRS / US-GAAP の 3 会計基準に対応。
            23 業種の ConceptSet を自動導出し、銀行・保険・証券・建設・鉄道の
            5 業種は専用科目マッピングを提供する。
            有報・四半期報告書・半期報告書など XBRL を含む全書類タイプで動作する。
            スレッドセーフではない。
        """
        from edinet.exceptions import EdinetAPIError

        if not self.has_xbrl:
            raise EdinetAPIError(
                0,
                f"書類に XBRL が含まれていません (doc_id={self.doc_id}, xbrlFlag=0)。",
            )
        resolved_taxonomy_path = self._resolve_taxonomy_path(taxonomy_path)
        xbrl_path, xbrl_bytes = self.fetch()
        return self._build_statements(resolved_taxonomy_path, xbrl_path, xbrl_bytes)

    async def axbrl(self, *, taxonomy_path: str | None = None) -> Statements:
        """XBRL を非同期で解析し財務諸表コンテナを返す。

        ``xbrl()`` の非同期版。ネットワーク I/O（ZIP ダウンロード）のみ
        非同期で、パース処理は同期的に実行する。

        Args:
            taxonomy_path: EDINET タクソノミのルートパス。
                省略時は ``configure(taxonomy_path=...)`` で設定された値を使用する。

        Returns:
            Statements コンテナ。

        Raises:
            EdinetAPIError: 当該書類に XBRL が無い場合。
            EdinetConfigError: taxonomy_path が未設定の場合。
            EdinetParseError: XBRL パースに失敗した場合。
            EdinetError: 通信層の失敗。

        Note:
            J-GAAP / IFRS / US-GAAP の 3 会計基準に対応。
            23 業種の ConceptSet を自動導出し、銀行・保険・証券・建設・鉄道の
            5 業種は専用科目マッピングを提供する。
            有報・四半期報告書・半期報告書など XBRL を含む全書類タイプで動作する。
            スレッドセーフではない。
        """
        from edinet.exceptions import EdinetAPIError

        if not self.has_xbrl:
            raise EdinetAPIError(
                0,
                f"書類に XBRL が含まれていません (doc_id={self.doc_id}, xbrlFlag=0)。",
            )
        resolved_taxonomy_path = self._resolve_taxonomy_path(taxonomy_path)
        xbrl_path, xbrl_bytes = await self.afetch()
        return self._build_statements(resolved_taxonomy_path, xbrl_path, xbrl_bytes)

    # --- ファクトリメソッド ---

    @classmethod
    def from_api_response(cls, data: dict[str, Any]) -> Filing:
        """EDINET API の書類一覧レスポンス（1件分の dict）から Filing を生成する。

        Args:
            data: get_documents() の results 配列の1要素。

        Returns:
            Filing オブジェクト。

        Raises:
            KeyError: 必須フィールド（docID, seqNumber, submitDateTime）が欠落。
            ValueError: 日時文字列のフォーマットが不正。
        """
        return cls(
            # === 必須フィールド（data["key"] — 欠落は KeyError で即座に検知） ===
            seq_number=data["seqNumber"],
            doc_id=data["docID"],
            doc_type_code=data.get("docTypeCode"),
            ordinance_code=data.get("ordinanceCode"),
            form_code=data.get("formCode"),
            filer_name=data.get("filerName"),
            submit_date_time=_parse_datetime(
                data["submitDateTime"], field_name="submitDateTime",
            ),
            doc_description=data.get("docDescription"),
            # === 任意フィールド（data.get() — None で保持、デフォルト空文字は使わない） ===
            edinet_code=data.get("edinetCode"),
            sec_code=data.get("secCode"),
            jcn=data.get("JCN"),
            fund_code=data.get("fundCode"),
            # 日付・期間
            period_start=_parse_date(data.get("periodStart")),
            period_end=_parse_date(data.get("periodEnd")),
            # 関連書類
            issuer_edinet_code=data.get("issuerEdinetCode"),
            subject_edinet_code=data.get("subjectEdinetCode"),
            subsidiary_edinet_code=data.get("subsidiaryEdinetCode"),
            current_report_reason=data.get("currentReportReason"),
            parent_doc_id=data.get("parentDocID"),
            ope_date_time=_parse_datetime_optional(
                data.get("opeDateTime"), field_name="opeDateTime",
            ),
            # ステータス
            # API 仕様上は常に存在する想定だが、防御的にデフォルト "0" を採用する。
            # Note: これにより API 仕様変更（フィールド削除）時の検知が遅れる可能性がある。
            # 必須化を検討する場合は KeyAccess に変更すること。
            withdrawal_status=data.get("withdrawalStatus", "0"),
            doc_info_edit_status=data.get("docInfoEditStatus", "0"),
            disclosure_status=data.get("disclosureStatus", "0"),
            # コンテンツフラグ
            has_xbrl=_parse_flag(data.get("xbrlFlag")),
            has_pdf=_parse_flag(data.get("pdfFlag")),
            has_attachment=_parse_flag(data.get("attachDocFlag")),
            has_english=_parse_flag(data.get("englishDocFlag")),
            has_csv=_parse_flag(data.get("csvFlag")),
            # 法定・任意
            legal_status=data.get("legalStatus", "0"),
        )

    @classmethod
    def from_api_list(cls, api_response: dict[str, Any]) -> list[Filing]:
        """get_documents() の戻り値全体から Filing リストを生成する。

        各要素の変換中に例外が発生した場合、何件目のどの書類で
        壊れたかを含む ValueError にラップして raise する。

        Args:
            api_response: get_documents() の戻り値（metadata + results を含む dict）。
                          引数名で「これは API のレスポンスそのもの」であることを明示。

        Returns:
            Filing オブジェクトのリスト。

        Raises:
            ValueError: 個別の Filing 変換に失敗した場合（元の例外を chain）。
        """
        results = api_response.get("results", [])
        filings: list[Filing] = []
        for i, doc in enumerate(results):
            try:
                filings.append(cls.from_api_response(doc))
            except (KeyError, TypeError, ValueError) as e:
                doc_id = doc.get("docID", "?") if isinstance(doc, dict) else "?"
                doc_type = doc.get("docTypeCode", "?") if isinstance(doc, dict) else "?"
                raise ValueError(
                    f"書類の変換に失敗しました (index={i}, "
                    f"docID={doc_id}, docTypeCode={doc_type}): "
                    f"keys={sorted(doc.keys()) if isinstance(doc, dict) else type(doc).__name__}"
                ) from e
        return filings

    # --- 表示 ---

    def __str__(self) -> str:
        """コンソールでの簡潔な表示。"""
        filer_name = self.filer_name or "(不明)"
        return f"Filing({self.doc_id} | {filer_name} | {self.doc_type_label_ja})"



from dataclasses import dataclass

__all__ = ['SummaryMapping', 'all_summary_mappings', 'lookup_summary', 'summary_concepts_for_standard']

@dataclass(frozen=True, slots=True)
class SummaryMapping:
    '''SummaryOfBusinessResults concept の CK マッピング。

    Attributes:
        concept: jpcrp_cor のローカル名（例: ``"NetSalesSummaryOfBusinessResults"``）。
        canonical_key: 正規化キー（例: ``"revenue"``）。
        standard: 会計基準識別子。
            ``"jgaap"`` / ``"ifrs"`` / ``"usgaap"`` / ``"jmis"``。
    '''
    concept: str
    canonical_key: str
    standard: str

def lookup_summary(concept: str) -> str | None:
    '''SummaryOfBusinessResults concept から CK を返す。

    Args:
        concept: jpcrp_cor のローカル名
            （例: ``"NetSalesSummaryOfBusinessResults"``）。

    Returns:
        正規化キー文字列。登録されていない concept の場合は ``None``。
    '''
def all_summary_mappings() -> tuple[SummaryMapping, ...]:
    """全 SummaryMapping を返す。

    Returns:
        全マッピングのタプル（J-GAAP → IFRS → US-GAAP → JMIS 順）。
    """
def summary_concepts_for_standard(standard: str) -> tuple[SummaryMapping, ...]:
    '''指定基準の SummaryMapping を返す。

    Args:
        standard: ``"jgaap"`` / ``"ifrs"`` / ``"usgaap"`` / ``"jmis"``。

    Returns:
        該当基準のマッピングタプル。未知の基準は空タプル。
    '''



"""脚注リンクパーサー。

XBRL インスタンス内の ``link:footnoteLink`` 要素をパースし、
Fact ID と脚注テキスト（※1, ※2 等）の紐付けを提供する。

J-GAAP の BS/PL で使用される脚注番号の解決に対応。
IFRS セクションでは footnoteLink は存在しない。

典型的な使い方::

    from edinet.xbrl.linkbase.footnotes import parse_footnote_links

    footnote_map = parse_footnote_links(parsed.footnote_links)
    notes = footnote_map.get("IdFact1234")
    for n in notes:
        print(n.text)  # "※1 減損損失の認識について..."
"""

from __future__ import annotations

import logging
import warnings
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass

from lxml import etree

from edinet.exceptions import EdinetParseError, EdinetWarning
from edinet.xbrl._namespaces import NS_LINK, NS_XLINK, NS_XML
from edinet.xbrl.parser import RawFootnoteLink

logger = logging.getLogger(__name__)

# ---------- Clark notation 定数 ----------
# footnoteLink は loc/footnote/arc の 3 種の要素で同一 XLink 属性を
# 反復参照するため、可読性と DRY の観点からモジュールレベル定数とする。

_TAG_LOC = f"{{{NS_LINK}}}loc"
_TAG_FOOTNOTE = f"{{{NS_LINK}}}footnote"
_TAG_FOOTNOTEARC = f"{{{NS_LINK}}}footnoteArc"

_ATTR_HREF = f"{{{NS_XLINK}}}href"
_ATTR_LABEL = f"{{{NS_XLINK}}}label"
_ATTR_ROLE = f"{{{NS_XLINK}}}role"
_ATTR_FROM = f"{{{NS_XLINK}}}from"
_ATTR_TO = f"{{{NS_XLINK}}}to"
_ATTR_ARCROLE = f"{{{NS_XLINK}}}arcrole"
_ATTR_LANG = f"{{{NS_XML}}}lang"

_ARCROLE_FACT_FOOTNOTE = "http://www.xbrl.org/2003/arcrole/fact-footnote"


# ========== データモデル ==========


@dataclass(frozen=True, slots=True)
class Footnote:
    """脚注。

    XBRL の ``link:footnote`` 要素から抽出した脚注情報。

    Attributes:
        label: xlink:label 属性値（アーク接続用識別子）。
        text: 脚注テキスト（例: ``"※1 当事業年度の…"``）。
        lang: xml:lang 属性値（例: ``"ja"``）。
        role: xlink:role 属性値。
    """

    label: str
    text: str
    lang: str
    role: str


@dataclass(frozen=True, slots=True)
class FootnoteMap:
    """Fact ID → 脚注のマッピング。

    ``parse_footnote_links()`` の返り値。
    Fact の id 属性値をキーに、紐付けられた脚注を検索できる。

    Attributes:
        _index: Fact ID → Footnote タプルの辞書（arc で紐付けられたもののみ）。
        _all: XML に存在する全 Footnote のタプル（arc による紐付けの有無に関わらず
              収集。重複除去済み。frozen dataclass のデフォルト ``__eq__`` による
              全フィールド一致で判定）。
    """

    _index: dict[str, tuple[Footnote, ...]]
    _all: tuple[Footnote, ...]

    def get(self, fact_id: str) -> tuple[Footnote, ...]:
        """Fact ID から脚注を取得する。

        Args:
            fact_id: Fact の id 属性値。

        Returns:
            紐付けられた Footnote のタプル。見つからなければ空タプル。
        """
        return self._index.get(fact_id, ())

    def has_footnotes(self, fact_id: str) -> bool:
        """Fact に脚注が存在するかを返す。

        Args:
            fact_id: Fact の id 属性値。
        """
        return fact_id in self._index

    def all_footnotes(self) -> tuple[Footnote, ...]:
        """全脚注を返す。"""
        return self._all

    @property
    def fact_ids(self) -> tuple[str, ...]:
        """脚注が紐付けられた Fact ID の一覧を返す。"""
        return tuple(self._index.keys())

    def __len__(self) -> int:
        """脚注が紐付けられた Fact の数を返す。"""
        return len(self._index)

    def __contains__(self, fact_id: object) -> bool:
        """Fact に脚注が存在するかの確認。"""
        return fact_id in self._index


# ========== 内部ヘルパー ==========


def _parse_single_link(
    root: etree._Element,
) -> tuple[
    dict[str, list[Footnote]],
    list[Footnote],
]:
    """1 つの footnoteLink 要素をパースする。

    Args:
        root: ``link:footnoteLink`` のルート要素。

    Returns:
        (fact_id → Footnote リスト, 全 Footnote リスト) のタプル。
    """
    # 1. loc 走査: label → Fact ID リスト（1:N）
    loc_map: dict[str, list[str]] = defaultdict(list)
    for loc_elem in root.findall(_TAG_LOC):
        label = loc_elem.get(_ATTR_LABEL)
        href = loc_elem.get(_ATTR_HREF)
        if not label or not href:
            continue

        if not href.startswith("#"):
            warnings.warn(
                f"脚注リンク: loc の href がフラグメント形式ではありません: '{href}'",
                EdinetWarning,
                stacklevel=3,
            )
            continue

        fact_id = href[1:]
        if not fact_id:
            warnings.warn(
                "脚注リンク: loc の href が '#' のみで Fact ID が空です",
                EdinetWarning,
                stacklevel=3,
            )
            continue

        loc_map[label].append(fact_id)

    # 2. footnote 走査: label → Footnote
    fn_map: dict[str, Footnote] = {}
    all_footnotes: list[Footnote] = []
    for fn_elem in root.findall(_TAG_FOOTNOTE):
        label = fn_elem.get(_ATTR_LABEL)
        if not label:
            continue

        role = fn_elem.get(_ATTR_ROLE)
        if not role:
            warnings.warn(
                "脚注リンク: footnote に xlink:role が未設定のためスキップします",
                EdinetWarning,
                stacklevel=3,
            )
            continue

        text = "".join(fn_elem.itertext())
        lang = fn_elem.get(_ATTR_LANG) or ""

        fn = Footnote(label=label, text=text, lang=lang, role=role)
        fn_map[label] = fn
        all_footnotes.append(fn)

    # 3. arc 走査: from(loc label) → to(footnote label)
    index: dict[str, list[Footnote]] = defaultdict(list)
    for arc_elem in root.findall(_TAG_FOOTNOTEARC):
        arcrole = arc_elem.get(_ATTR_ARCROLE)
        if arcrole != _ARCROLE_FACT_FOOTNOTE:
            continue

        from_label = arc_elem.get(_ATTR_FROM)
        to_label = arc_elem.get(_ATTR_TO)

        if not from_label or not to_label:
            warnings.warn(
                "脚注リンク: footnoteArc の from/to が未設定のためスキップします",
                EdinetWarning,
                stacklevel=3,
            )
            continue

        fact_ids = loc_map.get(from_label)
        if not fact_ids:
            warnings.warn(
                f"脚注リンク: loc ラベル '{from_label}' が見つかりません",
                EdinetWarning,
                stacklevel=3,
            )
            continue

        footnote = fn_map.get(to_label)
        if footnote is None:
            warnings.warn(
                f"脚注リンク: footnote ラベル '{to_label}' が見つかりません",
                EdinetWarning,
                stacklevel=3,
            )
            continue

        for fid in fact_ids:
            index[fid].append(footnote)

    logger.debug(
        "footnoteLink パース: loc=%d, footnote=%d, arc 解決=%d Fact",
        len(loc_map),
        len(fn_map),
        len(index),
    )

    return dict(index), all_footnotes


# ========== 公開 API ==========


def parse_footnote_links(
    raw_links: Sequence[RawFootnoteLink],
) -> FootnoteMap:
    """RawFootnoteLink 群から FootnoteMap を構築する。

    各 ``RawFootnoteLink.xml`` を再パースし、loc / footnote / footnoteArc の
    3 要素を抽出して Fact ID → Footnote のマッピングを構築する。

    複数の footnoteLink にまたがる脚注は統合される（同一 Fact ID に対する
    脚注はタプルにまとめられる）。

    Args:
        raw_links: ``ParsedXBRL.footnote_links``。

    Returns:
        構築された FootnoteMap。footnoteLink が空の場合は空の FootnoteMap を返す。

    Raises:
        EdinetParseError: XML パースに失敗した場合。
    """
    if not raw_links:
        return FootnoteMap(_index={}, _all=())

    merged_index: dict[str, list[Footnote]] = defaultdict(list)
    all_collected: list[Footnote] = []

    for raw in raw_links:
        try:
            root = etree.fromstring(raw.xml.encode("utf-8"))  # noqa: S320
        except etree.XMLSyntaxError as e:
            raise EdinetParseError(
                "脚注リンクの XML パースに失敗しました"
            ) from e

        index_part, all_part = _parse_single_link(root)

        for fact_id, footnotes in index_part.items():
            merged_index[fact_id].extend(footnotes)
        all_collected.extend(all_part)

    # _index: list → tuple
    final_index = {
        fid: tuple(fns) for fid, fns in merged_index.items()
    }

    # _all: 重複除去（frozen dataclass の __eq__/__hash__ で判定）
    final_all = tuple(dict.fromkeys(all_collected))

    logger.info(
        "脚注リンクのパース完了: %d Fact に脚注, 全 %d 脚注",
        len(final_index),
        len(final_all),
    )

    return FootnoteMap(_index=final_index, _all=final_all)



"""財務情報モデル定義。"""

from __future__ import annotations

import enum
import unicodedata
from collections.abc import Iterator
from dataclasses import dataclass, field
from decimal import Decimal
from pathlib import Path
from typing import TYPE_CHECKING, Final, Literal

if TYPE_CHECKING:
    import pandas as pd

    from edinet.xbrl.contexts import DimensionMember, Period
    from edinet.xbrl.taxonomy import LabelInfo
    from edinet.xbrl.taxonomy.concept_sets import ConceptSet

__all__ = ["LineItem", "StatementType", "FinancialStatement"]


@dataclass(frozen=True, slots=True, kw_only=True)
class LineItem:
    """型付き・ラベル付きの XBRL Fact。

    RawFact + StructuredContext + TaxonomyResolver から生成される、
    下流（Day 15 の FinancialStatement 組み立て等）が消費する主要データ型。

    Attributes:
        concept: Clark notation の QName（例: ``"{ns}NetSales"``）。
            RawFact.concept_qname をそのまま引き継ぐ。
        namespace_uri: 名前空間 URI。
        local_name: ローカル名（例: ``"NetSales"``）。
        label_ja: 日本語ラベル情報。TaxonomyResolver から取得。
        label_en: 英語ラベル情報。TaxonomyResolver から取得。
        value: 変換済みの値。数値 Fact は ``Decimal``、テキスト Fact は ``str``、
            nil Fact は ``None``。テキスト Fact の値は ``RawFact.value_raw``
            （``itertext()`` によるタグ除去済みプレーンテキスト）を使用する。
            HTML タグを含む原文が必要な場合は ``RawFact.value_raw`` を参照するか、
            ``edinet.xbrl.text`` パッケージの ``extract_text_blocks()`` を使用すること。
        unit_ref: unitRef 属性値。テキスト Fact は ``None``。
        decimals: decimals 属性値。``int`` / ``"INF"`` / ``None``。
        context_id: contextRef 属性値（トレーサビリティ用）。
        period: StructuredContext から転写した期間情報。
        entity_id: StructuredContext から転写した Entity ID。
        dimensions: StructuredContext から転写した Dimension 情報。
        is_nil: xsi:nil が真かどうか。
        source_line: 元 XML の行番号。
        order: 元文書内の出現順。
    """

    concept: str
    namespace_uri: str
    local_name: str
    label_ja: LabelInfo
    label_en: LabelInfo
    value: Decimal | str | None
    unit_ref: str | None
    decimals: int | Literal["INF"] | None
    context_id: str
    period: Period
    entity_id: str
    dimensions: tuple[DimensionMember, ...]
    is_nil: bool
    source_line: int | None
    order: int


class StatementType(enum.Enum):
    """財務諸表の種類。"""

    INCOME_STATEMENT = "income_statement"
    BALANCE_SHEET = "balance_sheet"
    CASH_FLOW_STATEMENT = "cash_flow_statement"
    STATEMENT_OF_CHANGES_IN_EQUITY = "statement_of_changes_in_equity"
    COMPREHENSIVE_INCOME = "comprehensive_income"


STATEMENT_TYPE_LABELS: Final[dict[StatementType, str]] = {
    StatementType.INCOME_STATEMENT: "損益計算書 (Income Statement)",
    StatementType.BALANCE_SHEET: "貸借対照表 (Balance Sheet)",
    StatementType.CASH_FLOW_STATEMENT: "キャッシュ・フロー計算書 (Cash Flow Statement)",
    StatementType.STATEMENT_OF_CHANGES_IN_EQUITY: "株主資本等変動計算書 (Statement of Changes in Equity)",
    StatementType.COMPREHENSIVE_INCOME: "包括利益計算書 (Statement of Comprehensive Income)",
}

_DATAFRAME_COLUMNS: Final[list[str]] = [
    "label_ja",
    "label_en",
    "value",
    "unit",
    "concept",
]


def _display_width(s: str) -> int:
    """文字列の端末表示幅を返す。

    CJK 全角文字は 2 カラム、それ以外は 1 カラムとして計算する。

    Args:
        s: 表示幅を計算する文字列。

    Returns:
        端末上の表示幅（カラム数）。
    """
    return sum(2 if unicodedata.east_asian_width(c) in "WF" else 1 for c in s)


def format_period(period: Period) -> str:
    """Period を表示用文字列に変換する。

    Args:
        period: 変換対象の期間オブジェクト。

    Returns:
        InstantPeriod → ``"2025-03-31"``、
        DurationPeriod → ``"2024-04-01〜2025-03-31"``。
    """
    from edinet.xbrl.contexts import DurationPeriod, InstantPeriod

    if isinstance(period, InstantPeriod):
        return str(period.instant)
    if isinstance(period, DurationPeriod):
        return f"{period.start_date}〜{period.end_date}"
    return str(period)


@dataclass(frozen=True, slots=True, kw_only=True)
class FinancialStatement:
    """組み立て済みの財務諸表。

    LineItem 群を選択ルール（期間・連結・dimension フィルタ）で
    絞り込み、JSON データファイルの order に従って並べたもの。

    ``items`` 内の各 ``LineItem.label_ja`` / ``LineItem.label_en`` は
    常に非 None（``LabelInfo`` 型）。TaxonomyResolver のフォールバック
    チェーン（T-7: 指定 role → 標準ラベル → 冗長ラベル → local_name）
    により保証される。

    Attributes:
        statement_type: 財務諸表の種類（PL / BS / CF）。
        period: この財務諸表が対象とする期間。BS の場合は
            ``InstantPeriod``、PL / CF の場合は ``DurationPeriod``。
            該当する Fact が 0 件の場合は ``None``。
        items: 並び順が確定した LineItem のタプル。
        consolidated: 連結（True）か個別（False）か。
        entity_id: Entity ID（トレーサビリティ用）。1 filing 内で
            entity_id は統一されるため単一値。空の statement
            （``items=()``）では空文字列 ``""``。
        warnings_issued: 組み立て中に発行された警告メッセージ一覧。
    """

    statement_type: StatementType
    period: Period | None
    items: tuple[LineItem, ...]
    consolidated: bool
    entity_id: str
    warnings_issued: tuple[str, ...]
    _concept_set: ConceptSet | None = field(default=None, compare=False, repr=False)
    _taxonomy_root: Path | None = field(default=None, compare=False, repr=False)

    def __getitem__(self, key: str) -> LineItem:
        """科目を日本語ラベル・英語ラベル・local_name で検索する。

        Args:
            key: 検索キー。以下の順で照合する:
                1. ``label_ja.text``（例: ``"売上高"``）
                2. ``label_en.text``（例: ``"Net sales"``）
                3. ``local_name``（例: ``"NetSales"``）

        Returns:
            最初にマッチした LineItem。同一ラベルの科目が
            複数存在する場合は、items 内で最初に見つかったものを返す。

        Raises:
            KeyError: マッチする科目が見つからない場合。
        """
        for item in self.items:
            if key in (item.label_ja.text, item.label_en.text, item.local_name):
                return item
        raise KeyError(key)

    def get(
        self, key: str, default: LineItem | None = None
    ) -> LineItem | None:
        """科目を検索する。見つからなければ default を返す。

        Args:
            key: 検索キー。照合順序は ``__getitem__`` と同一。
            default: マッチしない場合の返り値。

        Returns:
            マッチした LineItem、またはマッチしなければ default。
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key: object) -> bool:
        """科目の存在確認。``"売上高" in pl`` のように使う。"""
        if not isinstance(key, str):
            return False
        try:
            self[key]
        except KeyError:
            return False
        return True

    def __len__(self) -> int:
        """科目数を返す。"""
        return len(self.items)

    def __iter__(self) -> Iterator[LineItem]:
        """科目を順に返す。"""
        return iter(self.items)

    def __repr__(self) -> str:
        """簡潔な表現を返す。"""
        return (
            f"FinancialStatement(type={self.statement_type.value}, "
            f"period={self.period}, items={len(self.items)}, "
            f"consolidated={self.consolidated})"
        )

    def __str__(self) -> str:
        """プレーンテキスト表示。

        Rich なしでも ``print(pl)`` で読みやすいテーブルを出力する。
        科目名（日本語ラベル）と金額をカンマ区切り・右寄せで表示する。
        CJK 全角文字の表示幅を考慮して列を揃える。

        Note:
            value は ``build_statements()`` 経由で数値 Fact のみ含まれる前提。
            ``Decimal('Infinity')`` / ``Decimal('NaN')`` は来ない想定。
        """
        header = STATEMENT_TYPE_LABELS.get(
            self.statement_type, self.statement_type.value
        )
        scope = "連結" if self.consolidated else "個別"
        period_str = format_period(self.period) if self.period else "(期間なし)"

        lines: list[str] = [f"{header}  [{scope}] {period_str}", ""]

        if not self.items:
            lines.append("  (科目なし)")
            return "\n".join(lines)

        # ラベル幅を動的計算（最大 40 カラム幅でクリップ）
        max_width = min(
            max(
                (_display_width(item.label_ja.text) for item in self.items),
                default=0,
            ),
            40,
        )

        for item in self.items:
            label = item.label_ja.text
            label_width = _display_width(label)
            padding = max_width - label_width
            if isinstance(item.value, Decimal):
                # Note: 値幅は固定 20 カラム。混合精度の科目が入る場合は動的幅計算を検討
                value_str = f"{item.value:>20,}"
            elif item.value is None:
                value_str = f"{'—':>20}"
            else:
                value_str = f"{item.value!s:>20}"
            lines.append(f"  {label}{' ' * padding}  {value_str}")

        lines.append(f"\n  ({len(self.items)} 科目)")
        return "\n".join(lines)

    def to_dict(self) -> list[dict[str, object]]:
        """辞書のリストに変換する。

        pandas 不要で科目データを取得できる。
        LLM（RAG）パイプラインや軽量なデータ変換に適している。

        各辞書は ``to_dataframe()`` と同じ 5 キーを持つ:
        ``label_ja``, ``label_en``, ``value``, ``unit``, ``concept``。

        Returns:
            科目ごとの辞書のリスト。空 statement では空リスト。
        """
        return [
            {
                "label_ja": item.label_ja.text,
                "label_en": item.label_en.text,
                "value": item.value,
                "unit": item.unit_ref,
                "concept": item.local_name,
            }
            for item in self.items
        ]

    def to_dataframe(self, *, full: bool = False) -> pd.DataFrame:
        """pandas DataFrame に変換する。

        Args:
            full: True の場合、全カラム（context_id, period_type, dimensions 等）を
                含む DataFrame を返す。False（デフォルト）の場合は従来の 5 カラム。

        デフォルト（full=False）のカラム:

        - ``label_ja``: 日本語ラベル（``str``）
        - ``label_en``: 英語ラベル（``str``）
        - ``value``: 値（``Decimal | str | None``）
        - ``unit``: 単位（``str | None``）
        - ``concept``: concept のローカル名（``str``）

        ``value`` 列は ``Decimal`` のまま保持される（``object`` 型）。
        集計が必要な場合は ``df["value"].astype(float)`` で変換すること。

        返却される DataFrame の ``attrs`` に statement レベルのメタデータ
        （``statement_type``, ``consolidated``, ``period``, ``entity_id``）
        を付与する。

        Returns:
            科目名・金額等を含む DataFrame。

        Raises:
            ImportError: pandas がインストールされていない場合。
                ``pip install edinet[analysis]`` でインストールできる。

        Examples:
            >>> df = pl.to_dataframe()
            >>> df[df["value"] > 0]  # 正の値のみ抽出
            >>> df_full = pl.to_dataframe(full=True)  # 全カラム
        """
        if full:
            from edinet.dataframe.facts import line_items_to_dataframe

            return line_items_to_dataframe(
                self.items,
                metadata={
                    "statement_type": self.statement_type.value,
                    "consolidated": self.consolidated,
                    "period": format_period(self.period) if self.period else None,
                    "entity_id": self.entity_id,
                },
            )

        try:
            import pandas as pd_
        except ImportError:
            raise ImportError(
                "pandas is required for to_dataframe(). "
                "Install it with: pip install edinet[analysis]"
            ) from None

        rows: list[dict[str, object]] = []
        for item in self.items:
            rows.append({
                "label_ja": item.label_ja.text,
                "label_en": item.label_en.text,
                "value": item.value,
                "unit": item.unit_ref,
                "concept": item.local_name,
            })

        if not rows:
            df = pd_.DataFrame(columns=_DATAFRAME_COLUMNS)
        else:
            df = pd_.DataFrame(rows)

        df.attrs["statement_type"] = self.statement_type.value
        df.attrs["consolidated"] = self.consolidated
        df.attrs["period"] = format_period(self.period) if self.period else None
        df.attrs["entity_id"] = self.entity_id
        return df

    def to_csv(self, path: str | Path, **kwargs) -> None:
        """全カラム DataFrame を CSV に出力する。

        Args:
            path: 出力先ファイルパス。
            **kwargs: ``to_csv()`` に渡す追加引数。
        """
        from edinet.dataframe.export import to_csv

        to_csv(self.to_dataframe(full=True), path, **kwargs)

    def to_parquet(self, path: str | Path, **kwargs) -> None:
        """全カラム DataFrame を Parquet に出力する。

        Args:
            path: 出力先ファイルパス。
            **kwargs: ``to_parquet()`` に渡す追加引数。
        """
        from edinet.dataframe.export import to_parquet

        to_parquet(self.to_dataframe(full=True), path, **kwargs)

    def to_excel(self, path: str | Path, **kwargs) -> None:
        """全カラム DataFrame を Excel に出力する。

        Args:
            path: 出力先ファイルパス。
            **kwargs: ``to_excel()`` に渡す追加引数。
        """
        from edinet.dataframe.export import to_excel

        to_excel(self.to_dataframe(full=True), path, **kwargs)

    def _repr_html_(self) -> str:
        """Jupyter Notebook でのインライン HTML 表示。

        ``_concept_set`` が設定されている場合は階層表示を使用する。
        表示モジュールの import に失敗した場合は ``__repr__`` の
        プレーンテキスト HTML をフォールバックとして返す。
        """
        from html import escape as _escape

        try:
            from edinet.display.html import to_html

            if self._concept_set is not None:
                from edinet.display.statements import _resolve_abstract_labels

                labels = _resolve_abstract_labels(
                    self._concept_set, self._taxonomy_root,
                )
                return to_html(
                    self,
                    concept_set=self._concept_set,
                    abstract_labels=labels,
                )
            return to_html(self)
        except Exception:
            return f"<pre>{_escape(str(self))}</pre>"

    def __rich_console__(self, console, options):  # noqa: ANN001
        """Rich Console Protocol。

        ``from rich.console import Console; Console().print(pl)`` で
        フォーマットされたテーブルが表示される。

        ``_concept_set`` が設定されている場合は階層表示を使用する。
        rich がインストールされていない場合、この Protocol は呼ばれない
        （Rich Console 自体が存在しないため）。
        """
        try:
            if self._concept_set is not None:
                from edinet.display.statements import (
                    _resolve_abstract_labels,
                    render_hierarchical_statement,
                )

                labels = _resolve_abstract_labels(
                    self._concept_set, self._taxonomy_root,
                )
                yield render_hierarchical_statement(
                    self, self._concept_set, abstract_labels=labels,
                )
            else:
                from edinet.display.rich import render_statement

                yield render_statement(self)
        except ImportError:
            yield str(self)



from edinet.models.financial import FinancialStatement
from edinet.xbrl.taxonomy.concept_sets import ConceptSet

__all__ = ['to_html']

def to_html(statement: FinancialStatement, *, concept_set: ConceptSet | None = None, abstract_labels: dict[str, str] | None = None) -> str:
    """FinancialStatement を HTML テーブル文字列に変換する。

    ``display/statements.py`` の ``build_display_rows()`` を使用して
    ``DisplayRow`` リストを生成し、HTML テーブルに変換する。

    Args:
        statement: 表示対象の財務諸表。
        concept_set: 階層表示用の ConceptSet。None の場合はフラット表示。
        abstract_labels: 抽象概念のラベル辞書。

    Returns:
        HTML テーブル文字列。
    """
